/**
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * channel.h
 * (c) 2001-2007 Murat Deligonul
 */
#ifndef __CHANNEL_H
#define __CHANNEL_H

#include <map>
#include <string>
#include <utility>
#include <ostream>
#include "util/hash.h"
#include "util/stlutil.h"
#include "util/strings.h"
#include "config/types.h"
#include "logging/chatlog.h"
#include "addrentry.h"
#include "serverinfo.h"
#include "debug.h"

using util::strings::my_strdup;

/**
 * Represents an IRC channel and its users.
 * Currently we don't store the mode, it requires being aware of
 * server numerics and keeping track of mode changes and what not,
 * so we'll just do `/mode #chan' when reattaching...
 */
class channel {
public:
	typedef util::hash_table<const char *, std::pair<address_entry *, server_info::mode_flag_t>, 
					util::cstring_hash_fn, util::cstring_predicate> 	hash_table_t;
						
	enum chan_flag_t {
		NAMES_DATA_RECVED    =  1,
		WHO_REQ_SENT         = (1 << 1),
		RELAY_WHO_RESP       = (1 << 2), 
		WHO_DATA_RECVED      = (1 << 3),

		MODE_REQ_SENT		= (1 << 4),
		RELAY_MODE_RESP		= (1 << 5),
		MODE_DATA_RECVED	= (1 << 6)
	};

private:
	typedef 	std::pair<char, std::string> 		mode_entry;

	const char * 			_name;
	int 				flags;
	logging::chatlog * 		log;
	bool				logging;
	int 				log_options;

	/**
	  * Stores topic; pair stores who set it and when.
	  */
	std::string				topic_str;
	std::pair<std::string, std::string>	topic_data;

	/**
	  * Channel modes:
	  *		'D' modes do not have nickname parameters and are stored as numeric flags.
	  *		'B' and 'C' (valued) modes always have a parameter and are stored as pair
	  */
	server_info::mode_flag_t		modes;
	std::map<char, std::string>		valued_modes;

	/**
	  * Config system targets.
	  */
	config::target_list 		channel_target_list;
	static config::target_list	default_target_list;

	hash_table_t nick_hash;

public:
	channel(const server_info *, const char *, int);
	~channel();

	/** 
	 * The target list, used for accessing user preferences 
	 * for this channel. 
	 */
	const config::target_list& channel_target() const {
		return channel_target_list;
	}

	/**
	  * Default.
	  */
	static const config::target_list& default_channel_target() {
		return default_target_list;
	}

	/**
	  * Channel topic
	  */
	const char * topic() const {
		return topic_str.c_str();
	}
	void set_topic(const char * str) {
		topic_str = str;
	}
	bool has_topic() const {
		return !topic_str.empty();
	}
	const char * topic_setter() const {
		return topic_data.first.c_str();
	}
	const char * topic_time() const {
		return topic_data.second.c_str();
	}
	void set_topic_data(const char * setter, const char * time) {
		topic_data.first = setter;
		topic_data.second = time;
	}

	/**
	  * Channel flags.
	  * NOTE: these are not modes.
	  */
	int chan_flags() const { 
		return flags; 
	}
	bool is_chan_flag_set(chan_flag_t f) const {
		return (flags & f);
	}
	void set_chan_flag(chan_flag_t f) { 
		flags |= f; 
	}
	void clear_chan_flag(chan_flag_t f) { 
		flags &= ~f; 
	}
	void reset_chan_flags() {
		flags = 0;
	}
	/**
	  * Returns true if all channel data from the server has been obtained.
	  */
	bool has_all_data() const {
		return is_chan_flag_set(NAMES_DATA_RECVED) &&
			is_chan_flag_set(WHO_DATA_RECVED) &&
			is_chan_flag_set(MODE_DATA_RECVED);
	}

	/**
	  * Channel modes.
	  * B/C type modes (e.g. +k, +l)
	  */
	void set_valued_mode(char , const char * );

	// get value of a mode; NULL if not set.
	const char * get_valued_mode(char c) const {
		std::map<char, std::string>::const_iterator i = valued_modes.find(c);
		return (i != valued_modes.end()) ? (*i).second.c_str() : NULL;
	}
	bool is_valued_mode_set(char c) const {
		return util::contains_if(valued_modes.begin(), valued_modes.end(), 
					util::pair_key_predicate<char, std::string>( c));
	}
	void clear_valued_mode(char c) {
		valued_modes.erase(c);
	}
	void clear_valued_modes() {
		valued_modes.clear();
	}

	/**
	  * Other channel modes.
	  */
	int  chan_modes() const {
		return modes;
	}
	bool is_chan_mode_set(server_info::mode_flag_t flag) const {
		return (modes & flag);
	}
	void set_chan_mode(server_info::mode_flag_t flag) {
		modes |= flag;
	}
	void clear_chan_mode(server_info::mode_flag_t flag) {
		modes &= ~flag;
	}
	void reset_chan_modes() {
		modes = 0;
	}
	
	// Get a string containing all modes
	void get_all_modes(const server_info *, std::string &) const;


	
	/* logging */
	logging::chatlog * get_chatlog() const { 
		return log; 
	}
	void set_chatlog(logging::chatlog * c) { 
		log = c; 
		if (log != NULL) {
			log->set_options(log_options);
		}
	}

	int 	get_log_options() const { 
		return log_options; 
	}
	void	set_log_options(int i) { 
		log_options = i;
		if (log != NULL) {
			log->set_options(log_options);
		}
	}

	bool	is_logging() const { 
		return logging; 
	}
	void	set_logging(bool b) { 
		logging = b; 
	}

	/**
	  * Nickname access 
	  * FIXME: these perform hash lookups, even though we can probably assume the given pointer
	  *	    refers to the exact object in the hash table.
	  */
	void add_nick(address_entry * n, server_info::mode_flag_t f = 0) {
		DEBUG("channel::add_nick() [%s]: %s (w/ flags %u)\n", _name, n->nick(), f);
		assert(n);
		nick_hash.insert(n->nick(), std::make_pair(n, f));
	}

	void del_nick(const address_entry * n) {
		DEBUG("channel::del_nick() [%s]: %s\n", _name, n->nick());
		assert(nick_hash.contains(n->nick()));
		nick_hash.erase(n->nick());
	}

	int nick_flags(const address_entry * n) const {
		assert(nick_hash.contains(n->nick()));
		hash_table_t::const_iterator i = nick_hash.find(n->nick());
		return (*i).second.second;
	}

	void set_nick_flags(const address_entry *n, server_info::mode_flag_t f) {
		assert(nick_hash.contains(n->nick()));
		hash_table_t::iterator i = nick_hash.find(n->nick());
		server_info::mode_flag_t &target =  (*i).second.second;
		target |= f;
	}

	void clear_nick_flags(const address_entry *n, server_info::mode_flag_t f) {
		assert(nick_hash.contains(n->nick()));
		hash_table_t::iterator i = nick_hash.find(n->nick());
		server_info::mode_flag_t &target = (*i).second.second;
		target &= ~f;
	}

	address_entry * lookup(const address_entry * a) { 
		hash_table_t::iterator i = nick_hash.find(a->nick());
		if (i == nick_hash.end()) {
			return NULL;
		}
		return (*i).second.first;
	}
	
	address_entry * lookup_nick(const char * n) { 
		hash_table_t::iterator i = nick_hash.find(n);		
		return (i == nick_hash.end() ? NULL : (*i).second.first);
	}

	bool has_nick(const char * n) const {
		return nick_hash.contains(n);
	}

	bool has_nick(const address_entry * a) const {
		return nick_hash.contains(a->nick());
	}

	const char * name() const { 
		return _name; 
	}
	size_t population() const { 
		return nick_hash.size(); 
	}
	hash_table_t::iterator nicks_begin() {
		return nick_hash.begin();
	}
	hash_table_t::iterator nicks_end() {
		return nick_hash.end();
	}
	void nick_hash_stats(util::hash_stats * ht) const {
		nick_hash.stat(ht);
	}
			
private: 
	/* forbidden operations */
	channel(const channel &);
	channel & operator=(const channel&);
};

inline std::ostream& operator << (std::ostream& out, const channel& chan) {
	out << chan.name();
	return out;
}

inline std::ostream& operator << (std::ostream& out, const channel * chan) {
	out << chan->name();
	return out;
}

#endif
