/**
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * dcc.h -- Part of the ezbounce IRC proxy
 * (C) 1999-2003 Murat Deligonul
 */

#ifndef __DCC_H
#define __DCC_H

#include <string>
#include <ctime>
#include <sys/types.h>
#include "util/exception.h"
#include "util/idgenerator.h"
#include "util/tracked_object.h"
#include "util/expirable_object.h"
#include "io/filter_types.h"
#include "dcc_socket.h"
#include "file_entry.h"
#include "proxy.h"

class conn;
class dcc_offer;

struct dcc_exception_tag {};
typedef util::exception<dcc_exception_tag> dcc_exception;

/**
 * Base class for DCC objects.
 */
class dcc : public util::tracked_object<dcc>, public util::expirable_object<dcc> {
protected:
	enum dcc_stat_t {
		DCC_CLOSED    = 0,
		DCC_LISTENING = 1,
		DCC_ACCEPTED  = 2,
		DCC_CONNECTING = 4,
		DCC_CONNECTED  = 8,
		DCC_FINISHED   = 16
	};
	
	enum sock_event_t {
		SOCK_READABLE  = 2,
		SOCK_WRITEABLE  = 4,
		SOCK_CONNECTED = 8,
		SOCK_CLOSED    = 16
	};
	
	enum dcc_sock_role_t {
		LISTENER,
		SENDER,
		RECEIVER
	};

private:
	conn * _owner;
	static util::id_generator idgen;

	int stat;
	time_t timestamp;
	time_t start;
	unsigned int _id;

public:
	enum dcc_event_t {
		DCC_EVENT_SUCCESS = 0,
		DCC_SESSION_CLOSED = 1,
		DCC_PIPE_ESTABLISHED,
		DCC_PIPE_CLOSED,
		DCC_PIPE_FAILED,
		DCC_SEND_ESTABLISHED,
		DCC_SEND_COMPLETE,
		DCC_SEND_FAILED,
		DCC_GET_ESTABLISHED,
		DCC_GET_COMPLETE,
		DCC_GET_FAILED,
		DCC_TIMED_OUT,
		DCC_ERROR,
		DCC_CONNECTION_ACCEPTED,
		DCC_USER_KILL,
		DCC_EVENT_MAX
	};

	explicit dcc(conn *);
	virtual ~dcc() { 
		set_owner(NULL); 
		idgen.clear(_id); 
	}
	
	/** 
	 * Accessors 
	 */
	conn * owner() { return _owner; }
	unsigned int id() const { return _id; }
	int status() const { return stat; }
	time_t last_event() const { return timestamp; }
	time_t start_time() const { return start; }
	time_t age() const { return ircproxy::instance()->time() - start; }
	
	virtual int  info(char *, int) = 0;
	virtual int  info(conn *) = 0;
	
	/**
	 * expirable_object<T> interface
	 */
	virtual int  status_check(time_t, const void ** = NULL) const = 0 ;
	virtual int  die(int, const void *);

	static const char * strevent(dcc_event_t);
	static const char * strstat(dcc_stat_t);
	static int   lookup_by_id(conn *, unsigned int,  dcc ** , dcc_offer **);
	
	int     announce_dcc_event(dcc_event_t , const char * = "" );
	void    set_owner(conn * c);
    
protected:
	void  set_id(int i) {
		_id = i;
	}
	void  set_status(int s) {
		stat = s;
	}
	void  set_last_event(time_t t) {
		timestamp = t;
	}

	virtual int socket_event(int s, int events, int extra) = 0;
	friend class dcc_socket;
	friend class dcc_offer;

private:
	/* Forbidden operations */
	dcc(const dcc &);
	dcc& operator=(const dcc&);
};

/**
 * A DCC object with a socket that listens.
 */
class dcc_with_listener : public dcc {
protected:
	dcc_socket * listener;
	virtual int  socket_event(int, int, int) = 0;

public:
	dcc_with_listener(conn *, const net::ap_pair&);
	virtual ~dcc_with_listener();
	virtual int  info(char *, int);
	virtual int  info(conn *) = 0;
	
	virtual int status_check(time_t, const void ** = NULL) const;
	virtual int  die(int, const void *);

	unsigned short port() { return listener->local_port(); }
	const char * addr() { return listener->local_addr(); }
	int sock_family() { return listener->get_family(); }

private: /** Forbidden **/
	dcc_with_listener(const dcc_with_listener &);
	dcc_with_listener & operator = (const dcc_with_listener &);
};

/**
 * Used to send files over a DCC connection
 */
class dccsend : public dcc_with_listener {
private:
	dcc_socket * 			receiver;	
	file_entry * 			file;		/* file being sent */
	file_entry::reader_t * 		reader;
	unsigned int 			sent;

	io::filter_list			filters;
	
	int queue_next_packet();	
		
public:
	dccsend(conn *, file_entry *, const io::filter_list&, off_t, const net::ap_pair&); 
	virtual ~dccsend();
	virtual int  status_check(time_t, const void ** = NULL) const;
	virtual int  die(int, const void *);
	virtual int  info(char *, int);
	virtual int  info(conn *);

protected:
	virtual int  socket_event(int, int, int);

private:
	dccsend(const dccsend&);
	dccsend& operator = (const dccsend&);
};


/**
 * Used to proxy DCC connections.
 */
class dccpipe : public dcc_with_listener {
private:
	dcc_socket * sender;
	dcc_socket * receiver;

	net::ap_pair s_address;		// info about the sender
    
public:
	dccpipe(conn *, const net::ap_pair&, const net::ap_pair&);
	~dccpipe();
	virtual int info(char * , int);
	virtual int info(conn *);
	
	virtual int status_check(time_t, const void ** = NULL) const;
	virtual int die(int, const void *);

protected:
	virtual int socket_event(int, int, int);

private:
	dccpipe(const dccpipe&);
	dccpipe& operator = (const dccpipe&);
};

/**
 * Incoming DCC file transfers.
 */
class dccget : public dcc {
private:
	dcc_socket * 			sender;
	unsigned int 			expected;
	unsigned int 			received;
	file_entry * 	 		file;
	file_entry::writer_t *  	writer;

	io::filter_list			filters;

public:
	dccget(conn * owner, file_entry *, dcc_offer *);
	virtual ~dccget();
	virtual int info(char * , int);
	virtual int info(conn *);

	/**
	 * expirable_object<T> interface
	 */
	virtual int status_check(time_t, const void ** = NULL) const;
	virtual int die(int, const void *);

protected:
	virtual int socket_event(int, int, int);

private: /* forbidden operations */
	dccget(const dccget &);
	dccget& operator=(const dccget &);
};

#endif	/* __DCC_H */
