// file_entry.h
/**
  * The on-disk file entry format has changed since the 
  * initial version.  It is somewhat compatible with the previous version,
  * enough to allow easy conversion to the new format.  New layout:
  * 
  *   [  Creation Time : time_t          ]
  *   [  Reserved      : unsigned int    ]		
  *   [  Flags         : unsigned int    ]
  *   [  Reserved      : unsigned short  ]
  *   [  Padding Size  : unsigned short  ]
  *   [  Owner Size    : unsigned short  ]
  *   [  .. 4 more size fields of type unsigned short ]
  *   [ The string data follows ]
  *   [ 0 or more bytes of padding, as a result of deleting entries ]
  */

#ifndef __FILE_ENTRY_H
#define __FILE_ENTRY_H

#include <string>
#include <ostream>
#include <cstring>
#include "util/counted_object.h"
#include "util/tracked_object.h"
#include "util/managed_table.h"
#include "io/chain.h"
#include "debug.h"

class directory;

class file_entry : 
#ifdef __DEBUG__
	public util::tracked_object<file_entry> {
#else
	public util::counted_object<file_entry> {
#endif
public:
	struct header {
		time_t 		creation;
		unsigned short  padding_length;
		unsigned short  reserved0;
		unsigned int 	flags;
		unsigned int    reserved1;
		/** lengths include 1 byte for null character **/
		unsigned short  owner_length;
		unsigned short  from_length;
		unsigned short  misc_length;
		unsigned short  desc_length;
		unsigned short  name_length;
	};

	typedef io::output_chain	writer_t;
	typedef io::input_chain		reader_t;

public:
	enum flag_t {
		FE_CHAT_LOG		= 0x0002,		/* chat log */
		FE_WORLD_READABLE 	= 0x0004,		/* readable by all */
		FE_WORLD_WRITEABLE 	= 0x0008,		/* writeable by all */
		FE_DIRECTORY		= 0x0020,		/* is actually a directory */
		FE_EMPTY_ENTRY		= 0x0040,		/* entry refers to empty space; only used on disk */
		FE_PRIVATE		= 0x0080,		/* accessable by only the owner */
	};

	file_entry(directory *, const char *);
	file_entry(directory *, const header *, const char *);
	file_entry(const file_entry &);
	file_entry& operator= (const file_entry &);

#ifdef __DEBUG__
	~file_entry() {
		DEBUG("%s: [%p] destructed '%s'\n", __PRETTY_FUNCTION__, this, name());
	}
#endif
	
private:
	enum string_id_t {
		OWNER 	= 0,		/* user it belongs to */
		FROM,			/* where it originated from */
		MISC,
		DESC,			
		NAME,			
		MAX_STRING
	};


	util::managed_table<std::string, MAX_STRING>	strings;
	directory * 					_dir;
	time_t          				_creation;
	unsigned int					_flags;
	unsigned int					_usage_count;
	unsigned int    				_size;		/* NOTE: size is not stored in database; used only for ls */
	writer_t *					fwriter;
	
public:
	const char * 	misc() const { return strings.get(MISC).c_str(); }
	const char * 	owner() const { return strings.get(OWNER).c_str(); }
	const char * 	from() const { return strings.get(FROM).c_str(); }
	const char * 	desc() const { return strings.get(DESC).c_str(); }
	const char * 	name() const { return strings.get(NAME).c_str(); }
	unsigned int 	flags() const { return _flags; }
	unsigned int    usage_count() const { return _usage_count; }
	unsigned int    size() const { return _size; }
	time_t		creation() const { return _creation; }	
	directory * 	dir() const { return _dir; }
	bool  		is_writing() const { return fwriter != NULL; }
	writer_t	*writer() const { return fwriter; }

	void  set_owner(const char * val) { strings.set(OWNER, val); }
	void  set_desc(const char * val) { strings.set(DESC, val); }
	void  set_from(const char * val) { strings.set(FROM, val); }
	void  set_name(const char * val) { strings.set(NAME, val); }
	void  set_misc(const char * val) { strings.set(MISC, val); }
	void  set_flags(unsigned int f) { _flags = f; }
	void  set_creation(time_t t) { _creation = t; }
	void  set_size(unsigned int s) { _size = s; }
	void  inc_count() { ++_usage_count; }
	void  dec_count() { --_usage_count; }	

	unsigned int fill_header(header *) const;
	int   fill_data(char *) const;

	/**
	 * Open and close readers and writers.
	 */
	reader_t * open_reader(reader_t::openmode);
	writer_t * open_writer(writer_t::openmode);
	void close(reader_t *);
	void close(writer_t *);

	void update_size();

	/** error codes **/
	enum {
		ERR_BASE=-100,
		ERR_LOCKED=-101,
		ERR_OPEN=-102,
		ERR_MAX=-103
	};

	static const char * strerror(int err);			

public:
	/**
	  * Function objects for STL sorting algorithms
	  **/

	/** Normal sort: Alphabetical order, but directories come first. **/
	class sort_normal {
	public:
		bool operator()(const file_entry &f1, const file_entry &f2) const {
			bool dir1 = f1.flags() & FE_DIRECTORY;
			bool dir2 = f2.flags() & FE_DIRECTORY;
			if (dir1 && !dir2) {
				return true; 
			} 
			else if (!dir1 && dir2) {
				return false;
			} 
			return (strcmp(f1.name(), f2.name()) < 0);
		}
	};
};

#ifdef __DEBUG__
/**
 * Assist with pretty printing of file entry object lists.
 */
inline std::ostream& operator << (std::ostream& out, const file_entry * f)
{
	out << f->name() << '@' << (void *) f->dir();
	return out;
}
#endif

#endif

