/**
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * addrcache.cc	-- address/channel/nickname caching system
 * (c) 2003-2007 Murat Deligonul
 */

#include "autoconf.h"

#include <list>
#include <algorithm>
#include <functional>
#include <cstdlib>
#include <cstring>
#include <cstdio>
#include "util/stlutil.h"
#include "util/tokenizer.h"
#include "net/socket.h"
#include "addrcache.h"
#include "textline.h"
#include "serverinfo.h"

#include "debug.h"

using util::delete_ptr;
using util::apply_second;
using namespace util::strings;

address_cache::address_cache(const server_info * s, int size, int options) :
	nick_hash(size, s->nick_hash_func(), s->casecmp_func()),
	channel_hash(5, s->channel_hash_func(), s->casecmp_func()),
	server(s)
{ 
	nick_hash.set_rehash_trigger(0,3); 
}

address_cache::~address_cache()
{
	/* before destroying the hash tables, we must destroy
	 * each indiviual member */
	std::for_each(nick_hash.begin(), nick_hash.end(), apply_second<delete_ptr>());
	std::for_each(channel_hash.begin(), channel_hash.end(), apply_second<delete_ptr>());
}

int address_entry::create_addr(const char * n, const char * a)
{
	assert(n);
	int len = strlen(n);
	int total = len;
	char * new_addr;

	if (a) {
		total += strlen(a);
	}

	new_addr = new char[total+2];
	strcpy(new_addr, n);
	*(new_addr + len) = 0;
	*(new_addr + len + 1)  = 0;
	if (a) {
		strcpy(new_addr + len + 1, a);
	}
	
	/* takes care of updating our addr pointer as well */
	update_ref_nick_ptrs(new_addr);
/* 	delete[] addr; addr = new_addr;		 */
	return 0;
}

/**
 * go through all referenced channels, finding nickname entry in
 * hash table and updating pointer of old key to new one.
 */
int address_entry::update_ref_nick_ptrs(char * new_addr_ptr)
{
        int * flags = 0;
	int   i = 0;
	std::list<channel *>::iterator iter = refs.begin(), e = refs.end();

	DEBUG("acache::update_refs(): updating old channel pointers of '%s'->'%s' [%zd channels]\n", addr, new_addr_ptr, refs.size());

        if (!refs.empty()) {
	        flags = new int[refs.size()];
	        while (iter != e) {
			channel * c = *iter;
			flags[i++] = c->nick_flags(this);
		        c->del_nick(this);
			++iter;
                }
	}

	delete[] addr;
	addr = new_addr_ptr;

	i = 0;
	iter = refs.begin();
	for (; iter != e; ++iter) {
		(*iter)->add_nick(this, flags[i++]);
	}

	delete[] flags;
	return 0;
}

address_entry::~address_entry()
{
	/* destroy all references to us */
	DEBUG("address_entry::~address_entry()\n");
	std::for_each(refs.begin(), refs.end(), 
		std::bind2nd(std::mem_fun(&channel::del_nick), this));
	delete[] addr;
}

/**
 * Make the user 'join' a channel 
 */
int address_cache::add_channel(const char * chan)
{
	DEBUG("acache::add_channel(): [%p] %s\n", this, chan);
	if (channel_hash.contains(chan)) {
		assert(0);
		return -1;	/* failure */
	}
	channel * c = new channel(server, chan, 7);
	channel_hash.insert(c->name(), c);
	assert(channel_hash.contains(c->name()));
	return channel_hash.size();
}

/**
 * Make this user 'part' a channel.
 * We must remove all references to this channel. This means:
 * going thru channel's nick_hash and for each nickname:
 *	- removing reference to channel
 *	- if it's down to zero references, destroy it
 */
int address_cache::delete_channel(const char *chan)
{
	DEBUG("acache::delete_channel() [%p]: %s\n", this, chan);
	channel * c = lookup_channel(chan); 
	if (!c)	{
		assert(0);
		return -1;
	}
	channel_hash.erase(chan);

	/* remove all users' references to this channel */
	channel::hash_table_t::iterator i = c->nicks_begin(),
					 e = c->nicks_end(); 

	for ( ; i != e; ++i) {
		address_entry * e = (*i).second.first;

		e->del_chan(c);

		if (e->references().empty()) {
			DEBUG("Deleting stale address_entry %p\n", e);
			nick_hash.erase(e->nick());
			delete e;
		}
	}

	delete c;
	assert(!channel_hash.contains(chan));
	return channel_hash.size();
}

/**
 * Add a new nickname to a channel's user list.
 * Pre-conditions: channel must be in the channel hash
 */
int address_cache::add_nick(const char * n, const char * addr, const char * chan)
{
	channel * c = lookup_channel(chan);
	if (!c)	{
		assert(0);
		return -1;
	}

	if (c->has_nick(n)) {
		// Nickname already in the channel user list for some reason ??
                DEBUG("acache::add_nick(): IRC logic problem: user already contains on channel? (%s-->%s)\n", n, chan);
		return -1;
	}


	address_entry * e = lookup_nick(n);
	/*
	 * Have we already cached this user?
	 */
	if (!e) {
		e = new address_entry(n, addr);
		nick_hash.insert(e->nick(), e);
		//DEBUG("address_cache::add_nick() -- new entry: %s on %s\n", e->nick(), chan);
	}

	e->add_chan(c);
	c->add_nick(e);

	assert(c->has_nick(e->nick()));
	return c->population();
}


/**
 * Remove a nickname from a channel.
 * If channel is not given, then remove it from all channels
 *  and purge its entry
 */
int address_cache::delete_nick(const char * n, const char * chan)
{
	channel * c = chan ? lookup_channel(chan) : NULL;
	if (chan && !c) {
		assert(0);
		return -1;
	}

	/* ensure that the nickname exists  */
	address_entry * e = lookup_nick(n);
	if (!e) {
		DEBUG("acache::delete_nick() trying to remove non-existent %s from %s\n", n, chan);
		return -1;
	}

	if (c) {
		DEBUG("Removing %s from %s\n", e->nick(), c->name());
		c->del_nick(e);
		e->del_chan(c);
		assert(c->lookup_nick(e->nick()) == NULL);
	}
	if (!c || e->references().empty()) {
		/* remove from all channels (destroy it) */
		DEBUG("Removing %s from ALL channels\n", e->nick());
		nick_hash.erase(e->nick());
		delete e;
	}

	return 0;
}

/**
 * This is tricky: because of the optimization, all of the
 * entries in each channel's nickhash are pointing to the same
 * thing. So once the nickname is changed, all the references to it
 * must also be updated
 */
int address_cache::change_nick(const char * old, const char * _new)
{
	address_entry * e = lookup(old);
	if (!e) {
		// assert(0);
		DEBUG("address_cache::change_nick() %s->%s -- original not found!\n", old, _new);
		return -1;
	}
	nick_hash.erase(e->nick());		/* don't forget to remove! */

	if (e->change_nick(_new) < 0) {
		/* ok, something went wrong */
		assert(0);
		return -1;
	}

	nick_hash.insert(e->nick(), e);
	return 0;
}

/**
  * Bind a nick to its (newly discovered) address.
  */
int address_cache::bind_nick(const char * nick, const char * addr)
{
	address_entry * e = lookup(nick);
	if (!e) {
		assert(0);
		return -1;
	}
	nick_hash.erase(e->nick());
	if (e->bind(addr) < 0) {
		assert(0);
		return -1;
	}
	nick_hash.insert(e->nick(), e);
	return 0;
}

int address_cache::flag_operator(int param, const char * nick, const char * chan, server_info::mode_flag_t flags)
{
	address_entry * e = lookup(nick);
	channel * c = lookup_channel(chan);

	if (!e || !c) {
		assert(0);
		return -1;
	}

	switch (param) {
	case 0:		/* GET */
		return c->nick_flags(e);
	case 1:		/* SET */
		c->set_nick_flags(e, flags);
		return 0;
	case 2: 	/* CLEAR */
		c->clear_nick_flags(e, flags);
		return 0;
	}

	abort();
	return 0;	/* this should never be reached */
}

void address_cache::stats(struct util::hash_stats * nt, util::hash_stats * ct)
{
	if (nt) {
		nick_hash.stat(nt);
	}
	if (ct) {
		channel_hash.stat(ct);
	}
}

/**
 * FIXME: implement or remove
 */
int address_cache::purge()
{
	return 0;
}

/**
 * these two functions parse the following numerics' raw data:
 * 353 [begin NAMES parse and continue]
 * 366 [signifies that the NAMES block is completed]
 *
 * requirements:
 *  must be on the channel
 *
 * Raw Numeric looks like this:
 * :irc.avalonworks.ca 353 druglord @ #channel :+Xfig moofo fiend- Fizzter ryan[WIN] +shea_
 * NOTE: original RFC 1459 doesn't have '@|=|*' character proceeding nickname
 *
 * however, we expect input in the form:
 * nick1 +nick2 @nick3 @nick4 %nick5 nick6 etc
 */
int address_cache::parse_NAMES_block(const char * channel, const textline& args)
{
	::channel * c = lookup_channel(channel);
	/* make sure we are on the channel,
	 * and that we haven't already parsed all the NAMES data for it */
	if (!c || (c->chan_flags() & channel::NAMES_DATA_RECVED)) {
                return -1;
	}

	args.tokenize(textline::MAX);

	const char * nick;
	unsigned int count = 0;
	while ( (nick = args[count++]) != NULL ) {
		DEBUG("Trying nick %s\n", nick);
		server_info::mode_flag_t codes = 0;

		nick = server->strip_prefix_symbols(nick, &codes);
		DEBUG("Trying nick %s\n", nick);
		add_nick(nick, NULL, channel);
		set_nick_flags(nick, channel, codes);
		DEBUG("address_cache::parse_NAMES_block(): adding %s to %s with codes %u\n", nick, channel, codes);
	}
	return (int) count;
}


int address_cache::end_NAMES_block(const char * channel)
{
        ::channel * c = lookup_channel(channel);
	if (!c || (c->chan_flags() & channel::NAMES_DATA_RECVED)) {
	        return -1;
	}
        c->set_chan_flag(channel::NAMES_DATA_RECVED);
	return 0;
}

/* example syntax:
 * :irc.avalonworks.ca 352 druglord #quickbasic ~murat xxxx.address.com irc.avalonworks.ca druglord H :0 yes
 * :irc.avalonworks.ca 315 druglord #quickbasic :End of /WHO list.
 * --> We are given text after channel field
 *
 * TODO: doesn't check if nickname is already bound to an address in the cache
 *              (would save some cycles avoiding mem allocation & such)
 */
int address_cache::parse_WHO_block(const char * channel, const textline& args)
{
        ::channel * c = lookup_channel(channel);
	if (!c || (c->chan_flags() & channel::WHO_DATA_RECVED)) {
	        return -1;
	}

	if (args < 5) {
	        DEBUG("address_cache::parse_WHO_block() -- confused by bogus input\n");
		return -1;
	}

	const char * ident = args[0]; 
	const char * host  = args[1];
	const char * nick  = args[3]; 

	char * buff = NULL;
	my_asprintf(&buff, "%s@%s", ident, host);

        if (bind_nick(nick, buff) < 0) {
	        DEBUG("address_cache::parse_WHO_block() -- bind_nick() failed on %s\n", nick);
	        delete[] buff;
		return -1;
        }

	delete[] buff;
	return 0;
}

int address_cache::end_WHO_block(const char * channel)
{
        ::channel * c = lookup_channel(channel);
	if (!c || (c->chan_flags() & channel::WHO_DATA_RECVED)) {
	        return -1;
	}
        c->set_chan_flag(channel::WHO_DATA_RECVED);
        return 0;
}


/**
 * Parse a MODE line from IRC server
 * format of data: +/-modes [nick1] [xxxx] [xxx] [xxx] etc..A
 *                 where data[0]...data[n] refer to individual tokens
 */
int address_cache::parse_MODE_line(const char * chan, const textline& data)
{
	channel * c = lookup_channel(chan);
	if (!c) {
	        return -1;
	}

	const char *pos = no_leading(data[0]);
	char plus = *pos;
	int cur_arg = 1;
	int processed = 0;

	if (plus != '+' && plus != '-') {
	        DEBUG("address_cache::parse_MODE_line(%s) [%p]: Unknown mode string %s\n", chan, this, data[0]);
		return -1;
        }
	bool set = (plus == '+');

	++pos;
	for (; *pos; ++pos) {
		if (*pos == '+') {
			set = true;
			continue;
		}
		else if (*pos == '-') {
			set = false;
			continue;
		}

	        server_info::chan_mode_t mode_flag = server->chan_mode_type(*pos);
		const char * param = NULL;
				
	        switch (mode_flag) {		
		case  server_info::CHAN_MODE_A:
			// always a parameter, but we're not interested in these.
			++cur_arg;
			break;

		case  server_info::CHAN_MODE_B:		
			// always a parameter.
			// usually only channel mode +k
			param = data[cur_arg++];
			if (param == NULL) {
				DEBUG("address_cache::parse_MODE_line(): missing required parameter for mode '%c'\n", *pos);
				break;
			}
			if (set) {
				c->set_valued_mode(*pos, param);
			}
			else {
				c->clear_valued_mode(*pos);
			}
			++processed;
			break;

		case  server_info::CHAN_MODE_C:
			// parameter only when setting. 
			// example: +l
			if (set) {
				param = data[cur_arg++];
				if (param == NULL) {
					DEBUG("address_cache::parse_MODE_line(): missing required parameter for (type C) mode '%c'\n", *pos);
					break;
				}
				c->set_valued_mode(*pos, param);
			}
			else {
				c->clear_valued_mode(*pos);
			}
			++processed;
			break;

		case  server_info::CHAN_MODE_D:
		   {
			// no parameters.  common channel modes like +tns
			server_info::mode_flag_t flag = server->mode_to_flag(mode_flag, *pos);
			if (!flag) {
				DEBUG("address_cache::parse_MODE_line(): unknown mode '%c'!\n", *pos);
				break;
			}
			if (set) {
				c->set_chan_mode(flag);
			}
			else {
				c->clear_chan_mode(flag);
			}
			++processed;
			break;
		   }

		case server_info::CHAN_MODE_PREFIX:		   
		   {
			// +o, +v and the like. always a parameter.
			param = data[cur_arg++];
			if (param == NULL) {
				DEBUG("address_cache::parse_MODE_line(): no target nickname for mode '%c'!\n", *pos);
				break;
			}

			// validate mode
			server_info::mode_flag_t flag = server->mode_to_flag(mode_flag, *pos);
			if (!flag) {
				DEBUG("address_cache::parse_MODE_line(): unknown nickname mode '%c'!\n", *pos);
				break;
			}

			// validate target nickname
			address_entry * e = lookup(param);
			if (!e) {
				DEBUG("(!) address_cache::parse_MODE_line() [%p] -- Dunno what to do with %s on %s [mode line: %s]\n",
					   this, param, chan, data.all());
				break;
			}

			if (set) {
				c->set_nick_flags(e, flag);
			}
			else {
				c->clear_nick_flags(e, flag);
			}
			DEBUG("  ---> processed MODE change on %s for %s %c%d\n", chan, e->nick(), plus, flag);
			++processed;
			break;
		   }

		default:
			// shouldn't get here.
			DEBUG("(!) address_cache::parse_MODE_line() [%p] -- unexpected character '%c'\r\n", this, *pos);
			break;
		}
        }
	return processed;
}


/**
 * Restartable functions
 */
class acache_rfunc : public rfunc {
protected:
	net::socket * target;

	channel * chan;
	char * nick;

	int  written;
	int  current_idx;		/* where we stopped */
	int  current_flags;		/* misc data of current item */

	const server_info * server;

	address_cache * acache;
	address_entry * current;

	channel::hash_table_t::const_iterator i, e;

public:
	acache_rfunc(net::socket * t, channel * c,
			const char * n, const server_info * s) : 
		target(t), chan(c),
	  	nick(my_strdup(n)), 
		written(0),
	  	current_idx(0), 
		current_flags(0),
	  	server(s),
	  	current(0),
	  	i(c->nicks_begin()),
	  	e(c->nicks_end()) { }

	~acache_rfunc() {
		delete[] nick;
	}

	virtual int execute(int) = 0;

private:
	acache_rfunc(const acache_rfunc &);
	acache_rfunc & operator= (const acache_rfunc &);
};

class names_rfunc : public acache_rfunc {
public:
	names_rfunc(net::socket * t,channel * c, const char * n, const server_info * s)
		: acache_rfunc(t,c,n,s) { }

	virtual int execute(int arg);

private:
	char irc_buffer[511];	

};

class who_rfunc : public acache_rfunc {
public:
	who_rfunc(net::socket *t, channel *c, const char *n, const server_info * s)
		: acache_rfunc(t,c,n,s) { }
	virtual int execute (int arg);
};


rfunc * address_cache::create_rfunc(address_cache::rfunc_t t, net::socket * target,
					const char * chan, const char * n) 
{
	channel * c = lookup_channel(chan);
	assert(n != NULL);
	assert(chan != NULL);
	assert(target != NULL);

	if (!c) {
		return 0;
	}

	if (t == RFUNC_WHO) {
		return new who_rfunc(target, c, n, server);
	}
	else if (t == RFUNC_NAMES) {
		return new names_rfunc(target, c, n, server);
	}

	DEBUG("acache::create_rfunc() -- Requested for unknown type %d\n", t);
	abort();
	return 0;
}

 /** 
  ** Return values for restartable functions:
  ** 	0 - operation finished
  **	    iterator deleted and set to 0
  **	    other relevant fields set to zero
  **
  **    >0 - operation to be resumed later
  **	    [return value] represents current position
  **	    iterator, current, current_idx updated with current position
  **		NOTE: current_idx and return value are 1 based
  **
  **    <0 - fatal error, invalid arguments given or other unrecoverable error
  **	    data will not be modified
  */
int names_rfunc::execute(int)
{
	char flag_buffer[33];    /* to store '@+%' etc. */

	/* :irc.avalonworks.ca 353 druglord @ #channel :+Xfig moofo fiend- Fizzter ryan[WIN] +shea_ */
	/* FIXME: this totally disregards whether the channel is secret/private/public whatever
	 * (i.e. the @ or = character after the nickname) */

	while (current || i != e) {
	        if (!current) {
			current       = (*i).second.first;
		        current_flags = (*i).second.second; 
			++i;
			server->str_prefix_symbols(current_flags, flag_buffer);
			assert(current_flags ? *flag_buffer != 0 : true);
                }

		if (!written) {
		        /* print the header */
			written = snprintf(irc_buffer, sizeof(irc_buffer), ":%s 353 %s @ %s :",
			        server->name(),
				nick,
				chan->name());
		}

		const size_t needed = written + strlen(flag_buffer) + strlen(current->nick()) + 4;
		if (needed > target->buffer_availableO()) {
			/**
			 * No space available, so save state and 
			 * resume later.
			 */
			return current_idx+1;
		}

		if (needed >= sizeof(irc_buffer)) {
			/**
			 * Filled up an IRC text line. Queue what we have, create
			 * new IRC line in next iteration.
			 */
			DEBUG("writing: %s\n", irc_buffer);
			assert(irc_buffer[0] == ':');
			target->queue(irc_buffer, written);
			target->queue("\r\n", 2);
			written = 0;
			continue;
		}
		
		/**
		 * Insert the next nickname ... 
		 */
                written += snprintf(irc_buffer + written, 
				sizeof(irc_buffer) - written, "%s%s ",
		                flag_buffer,
		                current->nick());

		assert(written > 0);
		assert(written <= 510);
		++current_idx;
		current = NULL;
		current_flags = 0;
	}

	/** 
	 * All nicks should have been processed 
	 */
	if (written) {
		assert(irc_buffer[0] == ':');
	        target->queue(irc_buffer, written);
		target->queue("\r\n", 2);
        }

	/**
	 * Make sure there's room for numeric 366 data 
	 */
	const size_t LEN_366 = 30 + strlen(server->name()) + strlen(nick) + strlen(chan->name());

	if (target->buffer_availableO() < LEN_366) {
		DEBUG("Ack! No room for numeric 366 on channel %s\n", chan->name());
		current = NULL;
		written = 0;
		return current_idx + 1;		/* FIXME: is the +1 needed (or correct?) */
	}

	target->printfQ(":%s 366 %s %s :End of /NAMES list.\r\n",
				server->name(), nick, chan->name());
	current = 0;
	assert(chan->population() == (unsigned) current_idx);
	return 0;	/* DONE ! */
}

/* Like above but deals with WHO data instead,
 * with slight differences in the algorithm
 * FIXME:
 * better way to handle the my_strdup() scenario ?? */
/* >> :irc.Prison.NET 352 druglord0 #channel muratd stanchion.cs.utexas.edu irc.Prison.NET druglord0 H :0 mmm, donuts
   >> :irc.Prison.NET 352 druglord0 #channel you 209-122-233-22.s1832.apx2.sbo.ma.dialup.rcn.com                                        irc.servercentral.net asivsn H@ :2 uh. screw you
   >> :irc.Prison.NET 352 druglord0 #channel ~fetish ip68-2-139-54.ph.ph.cox.net irc.arcti.ca thekeepr^ H :6 Gray   would be the color if I had a heart
   >> :irc.Prison.NET 352 druglord0 #channel stuff c213-100-41-49.swipnet.se efnet.xs4all.nl SirGoblin H :4 lollerskates   <3
   >> :irc.Prison.NET 352 druglord0 #channel ~jookz 216-160-192-241.phnx.qwest.net irc.inet.tele.dk jookz H :5 jookz
   >> :irc.Prison.NET 352 druglord0 #channel jedi jagdsl-13.smig.net irc.blessed.net infoe G :4 Obi-Wan Kanblomi
*/

int who_rfunc::execute(int)
{
	char flag_buffer[server_info::MAX_FLAGS+1];    /* to store '@+%' etc. */
	char irc_buffer[511];    /* RFC1459 says 512 max, including CRLF;
				  * so 510 effective + 1 for NULL */

	const char * const dummy_server = "unknown.server";
	const char * const dummy_info =  "user information unknown";
	const char * const dummy_status = "H";
	const int dummy_hops = 1;

	while (current || i != e) {
	        if (!current) {
			current	      = (*i).second.first;
		        current_flags = (*i).second.second;
			++i;
			server->str_prefix_symbols(current_flags, flag_buffer);
			assert(current_flags ? *flag_buffer != 0 : true);
                }

		/* :irc.Prison.NET 352 druglord #channel ~jedi jagdsl-13.smig.net irc.blessed.net infoe G :4 Obi-Wan 			  *   Kanblomi */
		const char * ident = "unknown";
		const char * host = "unknown.host";
		char * tmp = my_strdup(current->address());
		if (tmp) {
			char * dummy = strchr(tmp, '@');
			if (dummy) {
				*dummy++ = 0;
				host = dummy;
			}
			ident = tmp;
		}

		written = snprintf(irc_buffer, sizeof(irc_buffer), ":%s 352 %s %s %s %s %s %s %s :%d %s",
			server->name(), nick, chan->name(),
			ident,
			host,
			dummy_server, 	  /* IRC Server he's on */
			current->nick(),  /* his nickname*/
			dummy_status,	  /* status, e.g. here,gone,IRC Operator ... */
			dummy_hops,	  /* the hop count */
			dummy_info	  /* user information */
		);
		delete[] tmp;

		/**
		 * NOTE: snprintf() returns not how many bytes were written,
		 * but how many bytes *would have been* written had buffer length limit
		 * not been reached
		 */
		if (written < 0 || written > 510) {
			written = 510;
		}

		if (unsigned (written + 2) > target->buffer_availableO()) {
			/* resume later */
			written = 0;
			return current_idx+1;
		}

		/**
		 * Ready to queue this line
		 */
		target->queue(irc_buffer, written);
		target->queue("\r\n", 2);
		written = 0;
		current = 0;
		current_flags = 0;
		++current_idx;
	}

	/** 
	 * Make sure there's room for numeric 315 data 
	 */
	/* >> :irc.Prison.NET 315 nickname #ezb :End of /WHO list. */
	unsigned LEN_315 = 28 + strlen(server->name()) + strlen(nick) + strlen(chan->name());
	if (target->buffer_availableO() < LEN_315) {
		DEBUG("Ack! No room for numeric 315 on channel %s\n", chan->name());
		current = NULL;
		return current_idx + 1;		/* FIXME: is the +1 needed (or correct?) */
	}

	target->printfQ(":%s 315 %s %s :End of /WHO list.\r\n",
				server->name(), nick, chan->name());
	current = NULL;
	assert(chan->population() == (unsigned) current_idx);
	return 0;	/* DONE ! */
}
