/**
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * dcc.cc -- Part of the ezbounce IRC proxy
 * (C) 1999-2008 Murat Deligonul
 *
 * ---
 * This handles the dcc proxying and sending for ezbounce. 
 * The dcc command interception and redirecting is done in the conn class
 * --
 * FIXME: die() can be called twice during shutdown situations.
 */

#include "autoconf.h"

#include <list>
#include <iterator>
#include <fstream>
#include <cerrno>
#include <cstdlib>
#include <ctime>
#include <sys/types.h>
#include <unistd.h>
#include "util/strings.h"
#include "util/idgenerator.h"
#include "config/proxy.h"
#include "io/event.h"
#include "net/error.h"
#include "dcc.h"
#include "dcc_offer.h"
#include "dcc_socket.h"
#include "file_entry.h"
#include "filelib.h"
#include "directory.h"
#include "user.h"
#include "conn.h"
#include "proxy.h"
#include "debug.h"

using std::ios;
using std::ios_base;
using std::istream;
using std::string;
using net::socket_exception;
using net::ap_pair;
using util::strings::timestamp_full;
using util::strings::my_strdup;

/** 
 * How big shall the buffer be expandable too?
 * If you keep needing to increase this it means 
 * I goofed up somewhere! 
 */
static const unsigned BUFFSIZE = 32720;

/**
 * Packetsize for sending files 
 * Ought to be runtime configurable
 */
static const unsigned PACKETSIZE = 4096;


/* static */ util::id_generator dcc::idgen(4);

/** dcc_socket **/
void dcc_socket::on_connect()
{
	assert(role() == dcc::SENDER);
	owner()->socket_event(role(), dcc::SOCK_CONNECTED, 0);
	net::socket::on_connect();
}

int dcc_socket::on_writeable()
{
	assert(role() != dcc::LISTENER);
	assert(role() == dcc::RECEIVER || role() == dcc::SENDER);
	owner()->socket_event(role(), dcc::SOCK_WRITEABLE, 0);
	return 0;
}

int dcc_socket::on_readable()
{
	switch (role()) {
	case dcc::LISTENER:
		/** 
		 * Connection accepted on socket ?
		 */
		owner()->socket_event(role(), dcc::SOCK_READABLE, 0);
		return 0;

	case dcc::RECEIVER:
	case dcc::SENDER:
		// otherwise, read into buffer
		if (this->read() < 0) {
			owner()->socket_event(role(), dcc::SOCK_CLOSED, 0);
		} 
		else {
			owner()->socket_event(role(), dcc::SOCK_READABLE, 0);
		}
	}
	return 0;
}

void dcc_socket::on_disconnect(int e)
{
	owner()->socket_event(role(), dcc::SOCK_CLOSED, e);
}

void dcc_socket::on_connect_fail(int e)
{
	assert(role() == dcc::SENDER);
	owner()->socket_event(role(), dcc::SOCK_CLOSED, e);
}

/** The base dcc constructor */
dcc::dcc(conn * o) 
{
	DEBUG("dcc::dcc() -- [%p]\n", this);
	stat = 0;
	_owner = NULL;

	set_owner(o);
	/* handle dcc limits */
	const unsigned short max_dccs = o->user()->config()->get<int>(user_perms::MAX_DCCS);
	if (o != NULL && max_dccs > 0 && o->dcc_count() >= max_dccs) 
	{
		o->dec_dcc_count();
		throw dcc_exception("Max DCC Count Exceeded");
	}

	_id = idgen.generate();
	timestamp = start = ircproxy::instance()->time();
}

/** 
 * FIXME: should the limit be enforced here? 
 */
void dcc::set_owner(conn * c)
{
	DEBUG("dcc::set_owner() [%p] %p ---> %p\n", this, _owner, c);
	if (_owner) {
		_owner->dec_dcc_count();
	}
	if (c) {
		c->inc_dcc_count();
	}
	_owner = c;        
}

/**
 * 'condemn' this DCC. notify owner if necessary.
 */
int dcc::die(int e, const void * str)
{
	announce_dcc_event((dcc_event_t) e, (const char *) str);
	printlog("Killing DCC [%u]: %s: %s\n", id(), strevent((dcc_event_t) e), str);
	stat = 0;

	set_owner(NULL);
	return 0;
}

int dcc::announce_dcc_event(dcc_event_t e, const char * str)
{
	if (owner()) {
		owner()->cprintf("DCC Event [%u]: %s: %s\r\n", id(), strevent(e), str);
		owner()->regulate_io_events();
	}
	printlog("DCC Event [%u]: %s: %s\n", id(), strevent(e), str);
	return 0;
}

const char * dcc::strstat(dcc_stat_t t)
{
	switch (t)
	{
	case DCC_CONNECTING:
		return "connecting to";    
	case DCC_CONNECTED:
		return "established";
	case DCC_FINISHED:
		return "finished";
	case DCC_LISTENING:
		return "listening";
	case DCC_CLOSED:
		return "closed";
	case DCC_ACCEPTED:
		return "accepted";
	}
	return "unknown state";
}

const char * dcc::strevent(dcc_event_t e)
{
	if (e < 0 || e >= DCC_EVENT_MAX) {
		return "Unknown event";
	}
	static const char * dcc_event_table[] = {
		"DCC event succesful",
		"DCC session closed",
		"DCC Proxy session established",
		"DCC Proxy session closed",
		"DCC Proxy session failed",
		"DCC Send established",
		"DCC Send completed",
		"DCC Send failed",
		"DCC Get established",
		"DCC Get completed",
		"DCC Get failed",
		"DCC session timed out",
		"DCC session failure",
		"DCC Connection accepted",
		"DCC Killed by user request"
	};
	return dcc_event_table[e];
}

/** 
 * Given a unified dcc/offer handle, find who/what it belongs to
 * success:
 *      return 1, the proper pointer is assigned, the other zero
 * failure:
 *      return 0
 */
int dcc::lookup_by_id(conn * owner, unsigned int id, dcc ** d, dcc_offer ** offer)
{
	const std::list<dcc *>& dcc_list = ircproxy::instance()->dcc_list();
	const std::list<dcc_offer *>& dcc_offer_list = ircproxy::instance()->dcc_offer_list();

	std::list<dcc *>::const_iterator dcci = dcc_list.begin(),	
					dcce  = dcc_list.end();
		
	std::list<dcc_offer *>::const_iterator oi = dcc_offer_list.begin(),
						oe = dcc_offer_list.end(); 
	assert(d && offer);

	while (dcci != dcce) {
		*d = *dcci;
		if ((*d)->owner() == owner && (*d)->id() == id) {
			*offer = NULL;
			return 1;
		}
		++dcci;
	}

	while (oi != oe) {
		*offer = *oi;
		if ((*offer)->owner() == owner && (*offer)->id() == id) {
			*d = NULL;
			return 1;
		}
		++oi;
	}
	*d = NULL;
	*offer = NULL;
	return 0;
}

/** dcc_with_listener **/
dcc_with_listener::dcc_with_listener(conn * o, const ap_pair& interface)
                                        : dcc(o)
{
	DEBUG("dcc_with_listener::dcc_with_listener() -- [%p]\n", this);
	/* set up the listener socket */
	listener = 0;
	try {
		const char * iface = interface.first.c_str();
		unsigned short port = interface.second;
		listener = new dcc_socket(this, LISTENER);
		set_status(DCC_LISTENING);
		if (port) { /* listen in port range if necessary */
			listener->listen(iface, port, 1);
		}
		else {
			const string& ports = ircproxy::instance()->config().get<string>(proxy_config::DCC_PORTS);
			listener->listen(iface, ports.c_str(), 1);        
		}
	} 
	catch (...) {
		delete listener;
		throw;
	}
}

int dcc_with_listener::die(int e, const void * str)
{
	delete listener;
	listener = NULL;
	return dcc::die(e, str);
}

dcc_with_listener::~dcc_with_listener()
{
	DEBUG("dcc_with_listener::~dcc_with_listener() -- [%p]\n", this);
	delete listener;
}


int dcc_with_listener::status_check(time_t now, const void **) const
{
	if ((status() & DCC_LISTENING) &&
			(now - last_event()) >= 90) {
		return DCC_TIMED_OUT;
	}
	if (!status()) {
		return -1;  /* dead */
	}
	return DCC_EVENT_SUCCESS;
}

int dcc_with_listener::info(char * buff, int l)
{
	if (buff && status() == DCC_LISTENING) {
		return snprintf(buff, l, "Listening on %s:%d", listener->local_addr(), listener->local_port());
	}
	return 0;
}

/**
 * DCC Send constructor:
 * If construction is successful, the dccsend object takes ownership of the file_entry
 * resource (and the reader it opens) and will automatically release them upon destruction.
 *
 * If constructor fails, an exception is thrown and the provided file_entry resource
 * is untouched.
 */
dccsend::dccsend(conn * o, file_entry * fe, 
			const io::filter_list& flist, 
			off_t offset,
			const ap_pair& interface) 
                        : dcc_with_listener(o, interface),
				receiver(NULL),
				file(fe),
				reader(NULL),
				sent(offset),
				filters(flist)
{
	DEBUG("dccsend::dccsend() -- [%p] (sending %s)\n", this, fe->name());
	assert(offset >= 0);

	// NOTE: the file is opened for IO later.
}

dccsend::~dccsend()
{
	delete receiver;
}

/*
 * Read the next packet and queue it for sending.
 * Return:
 * -1: Fatal error. Object will need to be killed
 * 0: Nothing to send, or could not send anything because of non-fatal error or
 *      something else
 * >0: # of bytes sent in this packet
 *
 * Does NOT close any files on error!
 */
int dccsend::queue_next_packet()
{
	static char buffer[PACKETSIZE + 1];
	istream& stream = reader->get_stream();
	
	stream.read(buffer, PACKETSIZE);
	size_t bytesread = stream.gcount();

	if (stream.bad() || (stream.fail() && !stream.eof())) {
		DEBUG("dccsend::send_next_packet(): Error in read: %s\n", strerror(errno));
		return -1;
	}

	return receiver->queue(buffer, bytesread);
}


int dccsend::socket_event(int r, int events, int extra)
{
	if (r == LISTENER) {
		assert(events == SOCK_READABLE);
		/** accept the connection **/
		try {
			receiver = new dcc_socket(listener->owner(), listener, RECEIVER);
			set_status(DCC_CONNECTED);
			listener->close();
			delete listener;
			listener = NULL;
			/* ack packets are ignored */
			receiver->set_events(io::EVENT_WRITE);

			/* try to open the file now; attach filters */
			reader = file->open_reader(ios::in | ios::binary);
			io::attach_filters(*reader, filters);
			reader->complete();
		} 
		catch (std::exception &e) {
			/** could not accept connection, or open reader **/
			die(DCC_SEND_FAILED, e.what());
			return 0;
		}

		/** reader open, socket ready, advance file pointer **/
		istream& stream = reader->get_raw_stream();
		stream.seekg(sent);
		if (stream.fail()) {
			/** note: die() takes care of freeing resources **/
			die(DCC_SEND_FAILED, "unable to seek file pointer");
			return 0;
		}

		announce_dcc_event(DCC_SEND_ESTABLISHED, "");
		return 0;
	}
	/** 
	 * Receiver events follow 
	 */	
	if (events == SOCK_READABLE) {
		/* ignore it all */
		receiver->clear();                    
	}	
	if (events == SOCK_WRITEABLE) {
		if (queue_next_packet() <= 0) {
			return die(DCC_SEND_FAILED, strerror(errno));
		}

		int flushed = receiver->flushO();
		if (flushed < 0) {
			return die(DCC_SEND_FAILED, net::strerror(flushed));
		}

		sent += flushed;
		if (file->size() <= sent) { 
			set_status(DCC_FINISHED);
			receiver->set_events(io::EVENT_READ);
			set_last_event(ircproxy::instance()->time());
		}
		DEBUG("dccsend: [%p] %u/%u bytes of '%s' written!\n", this, sent, file->size(), file->name());
	}
	if (events == SOCK_CLOSED) {
		if (status() & DCC_FINISHED) {
			return die(DCC_SEND_COMPLETE, "");
		}
		/** Premature disconnect **/
		return die(DCC_SEND_FAILED, net::strerror(extra));
	}
	return 1;
}

/**
 * Definition of idle:
 *  Waiting for connections and 90 secs passed
 * 		-OR-
 *  dcc send complete and 30 secs passed 
 */
int dccsend::status_check(time_t now, const void **) const
{
	const bool finished = status() & DCC_FINISHED;
	if (finished && (now - last_event() >= 30)) {
		return DCC_SEND_COMPLETE;
	}
	return dcc_with_listener::status_check(now);
}    
    
int  dccsend::die(int e, const void *str)
{
	/* close *our* sockets */
	delete receiver;

	/* close file descriptor and release file library resources */
	if (file != NULL) {
		if (reader != NULL) {
			file->close(reader);
		}
		ircproxy::instance()->filelib()->close(NULL, file);
	}

	receiver = NULL;
	file = NULL;
	reader = NULL;
	set_status(0);
	return dcc_with_listener::die(e, str);
}

int dccsend::info(char * buff, int l)
{
	if (!status()) {
		return 0;
	}
	if (buff) {
		if (status() == DCC_LISTENING) {
			return dcc_with_listener::info(buff, l);
		}
		const char * what = strstat((dcc_stat_t)status());
		if (receiver) {
			return snprintf(buff, l, "SEND %s %s:%d (%u/%u)", what,
					receiver->peer_addr(), receiver->peer_port(), sent, file->size());
		}
		return snprintf(buff, l, "SEND %s (%u/%u)", what, this->sent, file->size());
	}
	return 0;
}

int dccsend::info(conn * c)
{
	c->cprintf("FIXME\r\n");
	return 0;
}

/*
 * Create a dcc pipe object. Fill 'sender' with info given.
 * By default binds to default interface on random port, and
 * waits for the receiver to connect.
 * By supplying something non-null for host_info you can
 * control what port or interface. Also when function completes
 * you will know what IP address the sock was bound.
 *
 * On success: listen_port is assigned non-0 value.
 */
dccpipe::dccpipe(conn * o ,
		const ap_pair& sender_address,
		const ap_pair& listen_address)
                : dcc_with_listener(o, listen_address),
			sender(NULL),
			receiver(NULL),
			s_address(sender_address)
{
	/** 
	  * At this point the constructor for the base class has been called
	  * and we should have a socket setup, waiting for connections.
	  */
//	DEBUG("dccpipe::dccpipe(%p, %s, %d, %s, %d) [%p]\n",
//			o, sender_address, sender_port, listen_address, listen_port,
//			this);
	assert(!sender_address.first.empty());
	assert(sender_address.second);

//	s_address = my_strdup(sender_address);
//	s_port   = sender_port;
	/* wait for connections .... */
}
int dccpipe::status_check(time_t now, const void **) const
{
	return dcc_with_listener::status_check(now);
}

int dccpipe::die(int e, const void * s2)
{
	delete sender;
	delete receiver;
	sender = receiver = NULL;
	return dcc_with_listener::die(e, s2);
}

int dccpipe::info(conn *c)
{
	c->cprintf("FIXME\r\n");
	return 0;
}

int dccpipe::info(char * buff, int l)
{
	if (status() == DCC_LISTENING) {
		return dcc_with_listener::info(buff, l);
	}
	if (buff) {
		const char * what = strstat(dcc_stat_t(status()));
		if (sender && receiver) {
			return snprintf(buff, l, "PROXY %s [%s:%d]<-->[%s:%d]", what,
					sender->peer_addr(), sender->peer_port(),
					receiver->peer_addr(), receiver->peer_port());
		}
		return snprintf(buff, l, "PROXY (%s)", what);         
	}             
	return 0;
}
    
dccpipe::~dccpipe()
{
	// Nothing to do
}

int dccpipe::socket_event(int r, int event, int extra)
{
	DEBUG("dccpipe::socket_event(%d, %d, %d) [%p]\n", r, event, extra, this);
	if (r == LISTENER) {
		/* accept the connection */
		assert(event == SOCK_READABLE);
		/** accept the connection **/
		try {
			receiver = new dcc_socket(listener->owner(), listener, RECEIVER);
			set_status(DCC_ACCEPTED | DCC_CONNECTING);

			listener->close();
			delete listener;
			listener = 0;

			/* Now we need to connect to the sender */
			sender = new dcc_socket(receiver->owner(), SENDER);
			int r = sender->async_connect(s_address.first.c_str(), "", s_address.second);

			if (r < 0) {
				return die(DCC_PIPE_FAILED, net::strerror(r));
			}
			announce_dcc_event(DCC_CONNECTION_ACCEPTED, receiver->peer_addr());
			/* now just wait ... */

		} 
		catch (std::exception &e) {
			/** could not accept connection **/
			die(DCC_PIPE_FAILED, e.what());
		}
		return 0;        
	}
	else {
		if (event == SOCK_CONNECTED) {
			assert(r == SENDER);
			/** Proxy is now established **/
			char buffer[256];
			snprintf(buffer, 256, "[%s : %d] <----> [%s : %d]",
					receiver->peer_addr(), receiver->peer_port(),
					sender->peer_addr(), sender->peer_port());                 

			announce_dcc_event(DCC_PIPE_ESTABLISHED, buffer);

			set_status(DCC_CONNECTED);
			sender->set_events(io::EVENT_READ | io::EVENT_WRITE);
			return 1;
		}
		else if (event == SOCK_CLOSED) {
			if (status() & DCC_CONNECTING) {
				// Failed connection
				die(DCC_PIPE_FAILED, net::strerror(extra));
			}
			else {
				die(DCC_PIPE_CLOSED, net::strerror(extra));
			}
			return 0;                
		}
		else if (event == SOCK_READABLE) {
			int i = 0;
			/**
			 * Transfer to the output queue of the other end.
			 */
			if (r == SENDER) {
				i = sender->flush(receiver);
			}
			else {
				i = receiver->flush(sender);
			}

			if (i < 0) {
				return die(DCC_PIPE_FAILED, net::strerror(i));
			}
		}
		else if (event == SOCK_WRITEABLE) {
			int i = 0;
			/* flush what we can down the pipe */
			if (r == SENDER) {
				i = sender->flushO();
			}
			else {
				i = receiver->flushO();
			}

			if (i < 0) {
				return die(DCC_PIPE_FAILED, net::strerror(i));
			}
		}

		/* Now, balance the buffers */
		u_long snd = sender->obuff_size();
		u_long rcv = receiver->obuff_size();

		int snd_flags = 0, rcv_flags = 0;
		if (snd) {
			snd_flags |= io::EVENT_WRITE;
		}
		if (rcv) {
			rcv_flags |= io::EVENT_WRITE;
		}
		if (rcv < BUFFSIZE) {
			snd_flags |= io::EVENT_READ;
		}
		if (snd < BUFFSIZE) {
			rcv_flags |= io::EVENT_READ;
		}

		DEBUG("Sender queue: %lu Receiver: %lu Events: S: %d R: %d\n", snd, rcv,
				snd_flags, rcv_flags);
		sender->set_events(snd_flags);
		receiver->set_events(rcv_flags);
	}
	return 1;
}

/**
  * NOTE: if constructor throws exception, offer and entry are
  * 	  untouched (i.e. offer is still valid, and entry is still
  *	  open for access)
  *
  *	  Otherwise the object assumes ownership of the resource and frees
  *	  it (and closes the writer it opened) upon destruction.
  */
dccget::dccget(conn * owner, file_entry * entry, dcc_offer * offer) 
	: dcc(owner),
		sender(NULL),
		file(entry),
		writer(NULL)
{
	assert(offer != NULL);
	assert(offer->offset() < offer->size());
	assert(offer->from());

	/** Ready to connect **/
	sender = new dcc_socket(this, SENDER);
	int r = sender->async_connect(offer->address(), "", offer->port());
	if (r < 0) {
		/* failure */
		sender->close();
		delete sender;
		throw dcc_exception(net::strerror(r));
	}

	/** Create file **/
	try {
		writer = file->open_writer(ios::binary | ios::out | ios::app | ios::ate);
		writer->complete();
		std::ostream& stream = writer->get_raw_stream();

		stream.seekp(offer->offset(), ios_base::beg);
		if (stream.fail()) {
			file->close(writer);
			throw dcc_exception("unable to seek file pointer to requested position");
		}
	}
	catch (...) {
		sender->close();
		delete sender;
		throw;
	}

	/** steal the offer's id **/
	unsigned int temp;
	temp = offer->id();
	offer->set_id(id());
	set_id(temp);

	received = offer->offset();
	expected = offer->size();
	set_status(DCC_CONNECTING);
	/** success **/
}

dccget::~dccget()
{
	DEBUG("dccget::~dccget() [%p]\r\n", this);
	assert(!status());  /* we should never be destroyed directly. call die() */
	delete sender;
}

int dccget::socket_event(int r, int events, int extra)
{
	assert(r == SENDER);

	if (events == SOCK_CONNECTED)  {
		/** ready to receive data **/
		assert(status() != DCC_CONNECTED);
		set_status(DCC_CONNECTED);
		announce_dcc_event(DCC_GET_ESTABLISHED, "");
		sender->set_events(io::EVENT_READ);
	}
	else if (events == SOCK_WRITEABLE) {
		/* send the ack packet */
		assert(status() == DCC_CONNECTED);
		unsigned int bytes = htonl(received);
		sender->write(&bytes, sizeof(unsigned int));

		if (received >= expected)  {
			/* done */
			set_status(DCC_FINISHED);
			set_last_event(ircproxy::instance()->time());
			return die(DCC_GET_COMPLETE, "");
		}
		sender->set_events(io::EVENT_READ);
	}
	else if (events == SOCK_READABLE) {
		/* save the data, check for quota exceeded, etc. */
		assert(status() == DCC_CONNECTED);
		std::ostream& stream = writer->get_stream();
		int written = sender->flush(stream, sender->ibuff_size());
		if (written < 0) {
			return die(DCC_GET_FAILED, net::strerror(written));
		}

		received += written;
		sender->set_events(io::EVENT_READ | io::EVENT_WRITE);
		set_last_event(ircproxy::instance()->time());
	}
	else if (events == SOCK_CLOSED) {      
		if (status() & DCC_FINISHED) {
			return die(DCC_GET_COMPLETE, "");
		}
		return die(DCC_GET_FAILED, net::strerror(extra));
	}
	return 1;
}

int  dccget::die(int e, const void *str)
{
	/* close *our* socket */
	delete sender;

	if (file != NULL) {
		if (writer != NULL) {
			file->close(writer);
			writer = NULL;
		}
		ircproxy::instance()->filelib()->close(NULL, file);
	}
	
	sender = NULL;
	file = NULL;
	set_status(0);
	return dcc::die(e, str);
}

/** 
  * idle when:  
  *		stat == 0 (signifying death)
  *             or, nothing has happpened for 1 minute
  */
int dccget::status_check(time_t now, const void **) const
{
	if ((status() & DCC_CONNECTED) && (now - last_event()) >= 60) {
		return DCC_TIMED_OUT;
	}
	if (!status()) {
		return -1;
	}
	return 0;
}

int dccget::info(conn * c)
{
	if (!status()) {
		c->cprintf("This is a dead DCC\r\n");
	} 
	else {
		c->cprintf("From: %s %d\r\n", this->sender->peer_addr(), this->sender->peer_port());
		c->cprintf("File: %s\r\n", this->file->name());
		c->cprintf("Received: %u/%u (%.2f)\r\n", received, expected, (float) received / expected * 100);
	}
	return 0;
}

int dccget::info(char * buff, int l)
{
	if (buff) {
		const char * what = strstat(dcc_stat_t(status()));
		if (file) {
			return snprintf(buff, l, "GET (%s) %s (%ul/%ul)", what, file->name(),
					received, expected);
		}
		return snprintf(buff, l, "GET (%s) (unknown progress)", what);
	}

	return 0;
}

dcc_offer::dcc_offer(conn * owner, const char * type,
                    const char * from, const char * to,
                    const char * address,
                    unsigned short port,
                    const char * filename,
                    unsigned int size,unsigned int offset, int f) 
{
	assert(type);
	assert(!((f & INCOMING) && (f & OUTGOING)));
	assert((f & INCOMING) || (f & OUTGOING));
	_owner = NULL;
	_size = size;
	_offset = offset;
	_flags = f;
	_start = ircproxy::instance()->time();
	_port  = port;

	set_owner(owner);

	/** no illegal filenames!! **/
	if (filename != 0 && !directory::is_legal_name(filename)) {
		owner->dec_dcc_count();
		throw dcc_exception("Illegal character(s) in filename");
	}


	const unsigned short max_dccs = owner->user()->config()->get<int>(user_perms::MAX_DCCS);
	if (owner != NULL && max_dccs > 0 && owner->dcc_count() >= max_dccs) {
		owner->dec_dcc_count();
		throw dcc_exception("DCC Limit Exceeded");
	}

	_id = dcc::idgen.generate();
	data.set(FROM,from);
	data.set(TO, to);
	data.set(TYPE, type);
	data.set(ADDRESS, address);
	data.set(FILENAME, filename);

	DEBUG("dcc_offer::dcc_offer() [%p] New offer (...)\n", this);
}

dcc_offer::~dcc_offer() 
{
	set_owner(NULL); 
	dcc::idgen.clear(_id); 
}

/** 
 * An offer is idle when:
 * nothing happens after 5 minutes (set flags to 0) 
 */
int dcc_offer::status_check(time_t now, const void **) const
{
	if (flags() == DEAD) {
		return -1;
	}
	if (now - _start > 60 * 5) {
		if (_owner) {
			_owner->cprintf("DCC Offer %d has expired\r\n", id());
		}
		return 1;    
	}
	return 0;
}

int dcc_offer::die(int, const void *)
{
	_flags = DEAD;
	return 0;
}


int dcc_offer::info(char * buff, int l)
{
	if (_flags && buff) {
		const char * arrow = "-->";
		if (_flags & INCOMING) {
			arrow = "<--";
		}
		return snprintf(buff, l, "%s %s %s[%s:%d]-->%s (%s %d)\r\n",
				arrow,
				type(), from(), address(), port(), 
				to(), filename(), size());
	}
	return 0;
}

int dcc_offer::info(conn * c)
{
	string ts;
	timestamp_full(ts, start_time());
	c->cprintf("Type: %s\r\n", type());
	c->cprintf("Created: %s\r\n", ts.c_str());
	c->cprintf("%s ------> %s\r\n", from(), to());
	c->cprintf("Address: %s [port: %d]\r\n", address(), port());
	c->cprintf("Other: %s %d\r\n", filename(), size());
	return 0;
}

void dcc_offer::set_owner( conn * c )
{
	DEBUG("dcc_offer::set_owner() [%p] %p ---> %p\n", this, _owner, c);    
	if (_owner) {
		_owner->dec_dcc_count();
	} 
	if (c) {
		c->inc_dcc_count();
	}
	_owner = c;
}
