/***************************************************************************
 * Part of ezbounce
 * (c) 2006 Murat Deligonul
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 *  * 3/06 -- Completely rewritten.  Class renamed.  
 *	    Several new features including support for directories.
 *	    Several features ditched to rely on inherent OS support instead.
 *	    Database format is mostly backwards compatible. 
 *	    Database no longer fully loaded into memory.
 */

#include "autoconf.h"

#include <vector>
#include <string>
#include <algorithm>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <cerrno>
#include <climits>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

// FIXME: check for these includes
#include <sys/param.h>

#ifdef HAVE_SYS_SYSLIMITS_H
#include <sys/syslimits.h>
#endif
#include "util/tokenizer.h"
#include "util/strings.h"
#include "util/stlutil.h"
#include "filelib.h"
#include "file_entry.h"
#include "directory.h"
#include "user.h"
#include "debug.h"

// XXX XXX XXX XXX XXX XXX XXX
// TODO:
// XXX XXX XXX XXX XXX XXX XXX
// need better error message from directory class if looking up blank name
// implement import
// implement destruction methods
// create more tests
// automatically import files in a new directory?
// possible cleanup for directory? (removing _parent field)
// set rehash parameters for the hash tables
// can we suspend rehash during prunes?
// check permissions during directory traversal
// figure out when file_entry objects for directories are being cached and if this is
// 		correct. figure out when the proper time to delete them is.
// ls doesn't report all errors properly
// 

using std::string;
using namespace util::strings;

/**
  * Construct a new file library in the directory 'p'.
  * Will automatically convert an older database file to our
  *  current version.
  */  
file_library::file_library(const char * p) : 
	directories(13),
	root(NULL)
{
	// check if we can access it.
	if (::access(p, R_OK | W_OK | X_OK) != 0) {
		// complain
		throw flib_exception("Unable to access requested path", errno);
	}

	// is it a dir ? 
	struct stat st;
	memset(&st, 0, sizeof(st));
	stat(p, &st);
	if (!S_ISDIR(st.st_mode)) {
		// complain	
		throw flib_exception("Requested path is not a directory");
	}

	root = new directory(p);
	directories.insert("/", root);

	// brand new file library?
	if (!root->exists()) {
		DEBUG("file_library::file_library(): Detected new filesystem; creating root database\n");
		root->create();
	}

	int j = root->validate();
	if (j < 0) {
		/* errors */
		DEBUG("file_library::file_library(): Unable to validate root directory: %s\n", directory::strerror(j));
		delete root;
		throw flib_exception("Unable to validate root directory", j);	
	}
	
	/** setup 'root' user and file access key **/
	root_user = new userdef("(system)");
	root_user->config()->set<bool>(user_perms::IS_ADMIN, true);
	root_key = new flib_key(root_user, root); 

	/** setup ownership and basic directories if needed **/
	root->entry()->set_owner(root_user->name());
	root->entry()->set_flags(file_entry::FE_DIRECTORY | file_entry::FE_WORLD_READABLE);

	int err = 0;
	file_entry * fe = open(root_key, "/home", O_RDWR, &err);
	if (fe == NULL) {
		/** create /home directory **/
		DEBUG("file_library::file_library(): creating brand new /home/ directory\n");
		directory * d = mkdir(root_key, "/home", &err);
		if (d == 0) {
			delete root_key;
			delete root_user;
			delete root;
			throw flib_exception("Unable to create '/home' directory");
		}
		d->entry()->set_flags(file_entry::FE_WORLD_READABLE);
		assert(strcasecmp(d->entry()->owner(), "(system)") == 0);
		/** success **/
		return;	
	}

	if (!(fe->flags() & file_entry::FE_DIRECTORY)) {
		close(root_key, fe);
		delete root_key;
		delete root_user;
		delete root;
		throw flib_exception("'/home' must be a directory");		
	}

	/** success **/
	DEBUG("file_library::file_library(): detected existing /home directory\n");
	close(root_key, fe);	
}

/**
 * Destroy active directory objets one by one, starting 
 * with the leaf nodes.
 */
file_library::~file_library() 
{
	using std::for_each;
	using util::delete_ptr;
	using util::counted_object;

	delete root_key;
	delete root_user;

	std::vector<directory *> dirs;
	root->level_order_traverse(dirs);		
	assert(dirs.size() == directories.size());

#ifdef __DEBUG__
	string str;
	util::print_into(dirs.rbegin(), dirs.rend(), str);
	DEBUG("%s: destroying in this order: %s\n", __PRETTY_FUNCTION__, str.c_str());
#endif
	assert(counted_object<flib_key>::count() == 0);
	for_each(dirs.rbegin(), dirs.rend(), delete_ptr());
}

/**
  * Allocate and return a key object used for accessing the file library.
  * Also ensures that a home directory exists for the user. 
  * ALSO: creates dcc-in, dcc-out, and chat log directories
  */ 
flib_key * file_library::create_key(const userdef * u, int * err)
{
	resolved_path rp;
	char * buf = my_strdup2("/home/", u->name()); 
	*err = resolve_path(root_key, buf, &rp);
	directory * dir = rp.dir();	
	
	if (*err == 0) {
		dir = load_directory(root_key, rp.dir(), rp.name(), err);
	}
	
	if (*err < 0) {
		if (*err != directory::ERR_NOMATCH) {
			// unusual error -- usually database access problem or 
			// file already existing on disk
			return NULL;
		}

		/** 
		 * Didn't exist: could be first time accesing, so create
		 * his home directory 
		 */
		DEBUG("file_library::create_key(): Creating new home directory for %s\n", u->name());
		dir = mkdir(root_key, buf, err);
		if (dir == NULL) {
			/** Ok, this shouldn't happen at all. **/
			delete[] buf;
			return NULL;
		}		
		dir->entry()->set_owner(u->name());
		dir->entry()->set_flags(file_entry::FE_DIRECTORY | file_entry::FE_PRIVATE);
		save(root_key, dir->entry());
	}			
	delete[] buf;
	
	flib_key * key = new flib_key(u, dir);	
	DEBUG("file_library::create_key(): constructed key for '%s' in directory: '%s'\n", u->name(), dir->rel_path());
	return key;
}

/**
  * Access a file with requested mode.  Doesn't open() it, per se, but 
  * succeeds only if requested modes match (as listed in the database file).
  * Increaes usage count of found file entry.
  *
  * This will also open directories.  
  *
  * Returns a file_entry pointer which must be released with file_library::close()
  * Returns null on error with err set to the appropriate error code.  
  */
file_entry * file_library::open(const flib_key * key, const char * name,
					int mode, int * err)
{
	resolved_path rp;
	
	*err = resolve_path(key, name, &rp);
	if (*err < 0) {
		return NULL;
	}	

	directory * dir = rp.dir();
	if (dir == root && *rp.name() == 0) {
		/** filename was "/" **/
		return root->entry();		
	}
	file_entry * entry = dir->cache_lookup(rp.name());
	
	if (entry == NULL) {
		*err = dir->batch_start();
		if (*err < 0) {
			return NULL;
		}

		entry = dir->disk_lookup(rp.name(), err);		
		if (entry == NULL) {
			if (!(mode & O_CREAT)) {
				*err = ERR_DOESNT_EXIST;
				dir->batch_end();
				return NULL;
			}
			/**
			  * File creation requested: check 
			  * permissions and create here 
			  */
			int j = this->access(key, dir->entry(), mode);
			if (j < 0) {
				*err = ERR_ACCESS_WRITE;
				dir->batch_end();
				return NULL;
			}

			/** 
			 * Create on disk. It wasn't in the file library, so it had better not 
			 * exist on disk already.  We can check this with the O_EXCL flag.
			 */
			mode |= O_EXCL;
			char * buff = my_strdup3(dir->abs_path(), "/", rp.name()); 
			j = ::open(buff, mode, 0600);
			delete[] buff;
			
			if (j < 0) {				
				*err = errno == EEXIST ? ERR_EXISTS_ON_DISK : ERR_CREATION;
			       	dir->batch_end();
				return NULL;
			}
			::close(j);
			
			/** create new file_entry and save to disk. **/
			entry = new file_entry(dir, rp.name());
			entry->set_owner(key->username());
			entry->set_flags(file_entry::FE_PRIVATE);
			j = dir->write(entry, true);
			if (j < 0) {
				delete entry;
				*err = j;
				dir->batch_end();
				return NULL;				
			}			
		}
	       	else if (mode & (O_EXCL)) {
			/** Exclusive create requested, but file exists. **/
			delete entry;
			*err = ERR_EXISTS;
			dir->batch_end();
			return NULL;
		}
		dir->cache_insert(entry);
		dir->batch_end();
	}

	int r = this->access(key, entry, mode);
	if (r < 0) {
		*err = r;
		return NULL;
	}

	entry->inc_count();	
	entry->update_size();
	return entry;
}

/**
  * 'Closes' a file by decrementing usage count of its
  * file_entry. 
  */
int file_library::close(const flib_key *, file_entry * entry) 
{
	directory * dir = entry->dir();
	assert(dir);

	file_entry * orig = entry;
	entry = dir->cache_lookup(entry->name());

	if (entry == NULL || entry != orig) {
		return ERR_BADFILE;
	}
	assert(entry->usage_count() > 0);
	entry->dec_count();
	
	/** if count has reached 0, remove from cache **/
	if (entry->usage_count() == 0) {
		dir->cache_remove(entry->name());
		delete entry;
	}
	return 0;
}

/**
  * Save a file_entry to disk. 
  */
int file_library::save(const flib_key * key, file_entry * entry) 
{
	int r = access(key, entry, O_WRONLY);
	if (r < 0) {
		return r;
	}
	r = entry->dir()->write(entry);
	if (r < 0) {
		return r;
	}
	return 0;
}


/**
  * Check whether access is allowed.
  * FIXME: need better super-user conditions
  * FIXME: make sure secret file is only readable by owner. 
  */
int file_library::access(const flib_key * key, file_entry * entry, int mode) 
{
	bool uname_match = !userdef::namecmp(key->username(), entry->owner());

	bool need_read = (mode & O_RDONLY) || (mode & O_RDWR);
	bool need_write = (mode & O_WRONLY) || (mode & O_RDWR);

	bool writeable = (uname_match || (entry->flags() & entry->FE_WORLD_WRITEABLE));
	bool readable = (uname_match || (entry->flags() & entry->FE_WORLD_READABLE));

	bool super_user = (key->get_user()->config()->get<bool>(user_perms::IS_ADMIN));

	if (super_user) {
		return 0;
	}
		
	if (!uname_match && (entry->flags() & file_entry::FE_PRIVATE)) {
		return ERR_PRIVATE_FILE;
	}

	if (need_read && !readable) {
		return ERR_ACCESS_READ;
	}

	if (need_write && !writeable) {
		return ERR_ACCESS_WRITE;
	}
	
	return 0;
}

/**
 * Alternate version of access() that operates on a string path.
 */
int file_library::access(const flib_key * key, const char * name, int mode)
{
	int err;
	file_entry * fe = this->open(key, name, mode, &err);
	if (!fe) {
		return err;
	}
	// if we can succesfully open it, then access must be ok
	this->close(NULL, fe);
	return 0;
}


/**
  * Remove a file from the library. 
  */
int file_library::unlink(const flib_key * key, const char * name) 
{
	resolved_path rp;
	int err = resolve_path(key, name, &rp);
	if (err < 0) {
		return err;
	}		
	
	name = rp.name();
	directory * dir = rp.dir();
	file_entry * cached = dir->cache_lookup(name);
	if (cached != 0) {
		if (cached->usage_count() > 0) {
			return ERR_BUSY;
		}
		if (cached->flags() & file_entry::FE_DIRECTORY) {
			return ERR_DIRECTORY;
		}
	}

	// find entry and delete in single database access
	err = dir->batch_start();
	if (err) {		
		return err;	
	}

	file_entry * e = dir->disk_lookup(name, &err);
	if (e == 0) {		
		dir->batch_end();
		return err;
	}
	if (e->flags() & e->FE_DIRECTORY) {
		delete e;
		dir->batch_end();
		return ERR_DIRECTORY;
	}
	
	// HACK: ensure only owner or file system admin can erase
	e->set_flags(e->flags() & ~e->FE_WORLD_WRITEABLE);
	err = access(key, e, O_WRONLY);
	delete e;
	if (err < 0) {
		dir->batch_end();
		return err;
	}		
	
	// if was cached, remove from cache and deallocate
	if (cached != NULL) {
		dir->cache_remove(name);
		delete cached;
	}
	
	// finally, remove from disk
	dir->unlink(name, true);
	dir->batch_end();
	return 0;
}
	

/**
  * Change to requested directory.
  * Check permissions -- must have read access to the directory.
  * Update usage counts.
  * Return 0 on success.
  */
int file_library::chdir(flib_key * key, const char * path) 
{
	resolved_path rp;
	int err = resolve_path(key, path, &rp);
	if (err < 0) {
		return err;
	}

	directory * old_dir = key->dir();
	directory * new_dir = load_directory(key, rp.dir(), rp.name(), &err);

	if (new_dir == NULL) {
		return err;
	}
			
	DEBUG("file_library::chdir(): to '%s' (trying %s -> %s)\n", 
			path, old_dir->rel_path(), new_dir->rel_path());
	
	if (new_dir == old_dir) {
		return 0;	
	}

	/** FIXME: this may be unnecessary once load_directory() starts checking permissions **/
	err = access(key, new_dir->entry(), O_RDONLY);
	if (err < 0) {
		return err;
	}	
	key->set_dir(new_dir);
	return 0;
}

/**
 * Create a new directory.
 */
directory * file_library::mkdir(const flib_key * key, const char * name, int * err)
{
	resolved_path rp;
	int r = resolve_path(key, name, &rp); 
	if (r < 0) {
		*err = r;
		return NULL;
	}

	directory * parent = rp.dir();
	name = rp.name();

	r = parent->batch_start();
	if (r < 0) {
		*err = r;
		return NULL;
	}

	file_entry * e = parent->disk_lookup(name, err);
	if (e != NULL) {
		// already exists!
		*err = ERR_EXISTS;
		parent->batch_end();
		delete e;
		return NULL;
	}
	
	// check for other weird errors
	if (*err != directory::ERR_NOMATCH) {
		parent->batch_end();
		return NULL;
	}

	// access check
	r = access(key, parent->entry(), O_WRONLY);
	if (r < 0) {
		*err = r;
		parent->batch_end();
		return NULL;
	}

	// create it on disk
	char * fp = my_strdup3(parent->abs_path(), "/", name);
	r = ::mkdir(fp, 0711);
	delete[] fp;

	if (r < 0) {
		*err = (errno == EEXIST) ? ERR_EXISTS_ON_DISK : ERR_MKDIR_FAILURE;
		parent->batch_end();
		return NULL;
	}
		 
	e = new file_entry(parent, name);
	e->set_owner(key->username());
	e->set_flags(parent->entry()->flags() | file_entry::FE_DIRECTORY);
	e->set_creation(time(NULL));

	// create directory database first, then try to write to disk.
	directory * new_dir = new directory(parent, e);
	new_dir->create();
	r = new_dir->validate();
	if (r == 0) {
		r = parent->write(e, true); 
	}		
	if (r < 0) {
		*err = r;
		delete new_dir;
		delete e;
		parent->batch_end();
		return NULL;
	} 

	// success	
	directories.insert(new_dir->rel_path(), new_dir);
	parent->cache_insert(e);
	parent->batch_end();
	return new_dir;
}

/**
 * Attempt to remove a directory.  First deletes the database file and
 * removes it from the file library, then makes a call to rmdir().  
 */
int file_library::rmdir(const flib_key * key, const char * name)
{
	resolved_path rp;
	int err = resolve_path(key, name, &rp);
	if (err < 0) {
		return err;
	}
	directory * parent = rp.dir();
	name = rp.name();
	
	directory * dir = load_directory(key, parent, name, &err);
	if (dir == NULL) {
		return err;
	}
	
	// sanity checks 
	if (dir->usage_count() > 1 || !dir->is_empty()) {
		return ERR_BUSY;
	}
	if (dir->cache_num_open() > 0) {
		return ERR_OPEN_FILES;
	}
	// ownership check
	err = access(key, dir->entry(), O_WRONLY);
	if (err < 0) {
		return err;
	}

	err = parent->unlink(dir->name(), false);
	if (err < 0) {
		return err;
	}

	// save this ...
	std::string dir_name = dir->name();
	
	dir->cache_prune();
	dir->destroy();
	err = ::rmdir(dir->abs_path());
	
	/**
	  * Since the database file has been erased by now, a failed rmdir()
	  * due to other reasons is ignored.  The directory object is destroyed
	  * and removed from the cache. 
	  */
	directories.erase(dir->rel_path());
	delete dir;
  	  
	/**
	 * And remove file_entry from parent's cache if necessary 
	 * (should always be)...
	 */
	file_entry * fe = parent->cache_lookup(dir_name.c_str());
	if (fe != NULL) {
		assert(fe->usage_count() == 0);
		parent->cache_remove(dir_name.c_str());
		delete fe;
	}

	if (err < 0) {
		return ERR_RMDIR_FAILURE;
	}
	return 0;
}

/**
  * ls: still needs a bit of work
  */
int file_library::ls(const flib_key * key, const char * pattern, std::vector<file_entry>& results)
{
	resolved_path rp;
	int err = resolve_path(key, pattern, &rp); 
	if (err < 0) {
		return err;
	}

	pattern = rp.name();
	directory * dir = rp.dir();
	directory * possible = load_directory(key, dir, pattern, &err);
	if (possible != 0) {
		// FIXME: not the best error reporting 
		dir = possible;
		pattern = "*";
	}
	err = this->access(key, dir->entry(), O_RDONLY);
	if (err < 0) {
		return err;
	}
	
	err = dir->ls(pattern, results);
	if (err < 0) {
		return err;
	}
	return 0;	
}

/**
  * Creates a directory, creating any subdirectories that don't exist along the way.
  * Returns the new directory iff:
  * 	- it didn't already exist
  * 	- any previous directories along the way already existed or were successfully created.
  */
directory * file_library::mkdirs(const flib_key * key, const char * target, int * err)
{
	using std::string;
	resolved_path rp;

	// initial check -- at the very least clean up the path.
	*err = resolve_path(key, target, &rp); 
	if (*err < 0 && *err != directory::ERR_NOMATCH) {
		return NULL;
	}

	// This is a bit slow and naive; start from the base of the path and keep
	// trying to create directories.
	std::vector<string> tokens;
	tokenize(rp.full_path(), "/", tokens);

	std::vector<string>::const_iterator i = tokens.begin(),
						e = tokens.end();

	// start at "/" and keep appending tokens
	string path = "";
	directory * d = NULL;
	*err = ERR_EXISTS;
	for (; i != e; ++i) {
		int err2 = 0;
		path += '/';
		path += *i;
		d = mkdir(key, path.c_str(), &err2);
		if (d == NULL && err2 != ERR_EXISTS) {
			DEBUG("file_library::mkdirs('%s'): bailing out with: %s\n", target, strerror(err2));
			*err = err2;
			return NULL;
		}
	}
	DEBUG("file_library::mkdirs('%s'): returning '%s'\n", target, path.c_str());
	return d;
}

/**
  * Given a parent directory, returns the 
  * directory object associated with name. 
  * (NOTE: name must not be terminated with a slash)
  */
directory * file_library::load_directory(const flib_key * key, directory * parent, 
						const char * name, int * err)
{
	const char * slash = (parent == root) ? "" : "/";
	
	/** Check if directory is already in cache **/
	char * full_path = my_strdup3(parent->rel_path(), slash, name);
	hash_table_t::iterator i = directories.find(full_path);
	delete[] full_path;
	if (i != directories.end()) {
		return (*i).second;
	}

	file_entry * entry = parent->cache_lookup(name);
	if (entry == NULL) {
		entry = parent->disk_lookup(name, err);
		if (entry == NULL) {
			/** not on disk, or other error **/
			/** (err set by disk_lookup() call) **/
			return NULL;
		}
		parent->cache_insert(entry);
	}
	if (!(entry->flags() & file_entry::FE_DIRECTORY)) {
		// not a directory
		*err = ERR_NOT_DIRECTORY;
		return NULL;
	}

	/** FIXME: check access permissions? **/

	directory * result = new directory(parent, entry);
	*err = result->validate();
	if (*err < 0) {
		delete result;
		return NULL;
	}
	directories.insert(result->rel_path(), result);		
	return result;
}

int file_library::resolve_path(const flib_key * key, const char * name, resolved_path * rp)
{
	/** temporarily set cwd to filesystem root **/
	int old_dir = ::open(".", O_RDONLY); 
	char * resolved = rp->buffer();
	::chdir(root->abs_path());
	int err = fixup_path(key, name, resolved);
	fchdir(old_dir);
	::close(old_dir);
	
	if (err < 0) {	
		// error
		return err;
	}

	DEBUG("file_library::resolve_path(): fixup returned: %s --> %s\n", name, resolved);
	assert(strchr(resolved, '/') != 0);
	
	// get last slash 
	char * slash = strrchr(resolved, '/');
	rp->name_ptr = slash + 1;
	
	// check for "/" case
	if (slash == resolved) {
		rp->_dir = root;
		return 0;
	}
	// temporarily cut string at last slash
	*slash = 0;
		
	hash_table_t::const_iterator i = directories.find(resolved);
	if (i != directories.end()) {
		rp->_dir = (*i).second;
		*slash = '/';
		return 0;
	}

	// directory was not found in the cache, and there's no guarantee 
	// that it is even in the file library
	tokenizer<> tok(resolved, resolved+strlen(resolved), delimeter_seeker<>("/"));

	err = 0;
	directory * pwd = root;
	for (tokenizer<>::const_iterator j = tok.begin(), e = tok.end();
					j != e;
					++j) { 
		const string& current = *j;
		pwd = load_directory(key, pwd, current.c_str(), &err);
		if (pwd == NULL) {
			*slash = '/';
			return err;
		}
	}
	
	*slash = '/';
	rp->_dir = pwd;
	return 0;		
}

/**
  * Determines the true path provided (i.e. resolving '.' and '..')
  * Removes extra '/' characters.
  * Ensures that it does not attempt to leave virtual root directory  
  * or follow symbolic links.
  * The returned path is guaranteed not to have a '/' at the end, unless
  *  it is the full path "/"
  * NOTE: call stat() and access() and thus assumes current working directory is 
  * file system's on-disk root directory.
  *  	rel_cwd   -- current directory, relative to virtual root of file library
  *	p 	  -- requested pathname to resolve
  *     resolved  -- buffer to store result in. must be >= PATH_MAX.
  *
  * return: 
  *		  0 on success
  *		< 0 otherwise
  */
/* static */ int file_library::fixup_path(const flib_key * key, const char * p, char * resolved)
{
	char * path = new char[PATH_MAX+1];
	/** Check for home directory access character **/
	if (p[0] == '~') {
		snprintf(path, PATH_MAX, "/home/%s/%s", key->username(), p+1);
	} 
	/** Other absolute or relative path **/
	else {
		const char * rel_cwd = (p[0] == '/') ? "" : key->cwd->rel_path();
		snprintf(path, PATH_MAX, "%s/%s", rel_cwd, p);
	}

	p = path;
	*resolved = 0;
	int s = 0;	

	do {
		switch (s) {
		case 0:
			if (*p == '.') {
				s = 1;
			}
			else if (*p != 0 && *p != '/') {
				strcat(resolved, "/");
				s = 3;
			}
			break;

		/** Got '.' **/
		case 1:
			if (*p == '.') {
				s = 2;
			}
			else if (*p == '/') {
				s = 0;
			}
			else if (*p != 0) {
				/** file names like ".file" **/
				strcat(resolved, "/.");
				s = 3;
			}
			break;

		/** Got ".." **/
		case 2:
			if (*p == '/' || *p == 0) {
				char *beg = strrchr (resolved, '/');
				if (beg != 0)
					*beg = 0;
				s = 0;
			} 
			else {
				/** messed up file names like "..file" **/
				strcat(resolved, "/..");
				s = 3;
			}
			break;

		/** Got potential directory name **/
		case 3:
			const char *end = strchr(p, '/');
			if (end == 0) { 
				end = p + strlen(p);
			}
			strncat (resolved, p - 1, end - p + 1);
			/** make sure it's not a symlink **/
			struct stat st;
			memset(&st, 0, sizeof(st));
			int r = lstat(resolved+1, &st);
			if (r >= 0 && S_ISLNK(st.st_mode)) {
				delete[] path;
				return ERR_SYMLINK;			
			}
			p = end;
			s = 0;
		}
	} while (*p++);	
	delete[] path;

	if (*resolved == 0) {
		strcpy(resolved, "/");
	}

	assert(strcmp(resolved, "/") == 0 || *(strrchr(resolved, '/')+1) != 0);
	return 0;
}

/**
 * Translate a file library error code.
 */
/* static */ const char * file_library::strerror(int e)
{
	switch (e) {
	case ERR_BADFILE:
		return	"Unknown file pointer";
	case ERR_CREATION:
		return	"Unable to create file";
	case ERR_PRIVATE_FILE:
		return	"Permission denied: private file";
	case ERR_ACCESS:
		return	"Permission denied";
	case ERR_ACCESS_READ:
		return	"Read access denied";
	case ERR_ACCESS_WRITE:
		return	"Write access denied";
	case ERR_DIRECTORY:
		return	"File is a directory";
	case ERR_NOT_DIRECTORY:
		return	"File is not a directory";
	case ERR_BUSY:
		return	"File or directory is busy";
	case ERR_FAILURE:
		return	"Operation failed";
	case ERR_OPEN_FILES:
		return	"Directory has open files";
	case ERR_RMDIR_FAILURE:
		return	"Unable to remove directory from disk";
	case ERR_MKDIR_FAILURE:
		return  "Unable to create directory on disk";
	case ERR_SYMLINK:
		return	"Path contains a symlink";
	case ERR_EXISTS:
		return	"File exists";
	case ERR_DOESNT_EXIST:
		return	"File not found";
	case ERR_EXISTS_ON_DISK:
		return  "File not in database, but already exists on disk";
	case ERR_CONFIG:
		return	"You don't have permission to use the file library";
	case ERR_NO_KEY:
		return	"You don't have a file library access key";
	case ERR_NO_FILELIB:
		return	"The file library doesn't exist; or failed to initialize";
	}
	return directory::strerror(e);
}


flib_key::flib_key(const userdef * u, directory * d) 
		: user(u), cwd(d)
{
	DEBUG("flib_key::flib_key(): Creating [global count: %zu]..\n", count());
        assert(cwd != NULL);
	cwd->inc_count();
}

flib_key::flib_key(const flib_key& that) 
              : user(that.user), cwd(that.cwd)
{
	DEBUG("flib_key::flib_key(): Copying [global count: %zu]..\n", count());
        assert(cwd != NULL);
        cwd->inc_count();
}

flib_key::~flib_key() 
{
	DEBUG("flib_key::~flib_key(): Destructing [global count: %zu]...\n", count()-1);
        assert(cwd->usage_count() > 0);
        cwd->dec_count();
}

void flib_key::set_dir(directory * new_dir)
{	
	assert(cwd->usage_count() > 0);	
        assert(new_dir != NULL);
	new_dir->inc_count();
	cwd->dec_count();
        cwd = new_dir;
}

/** FIXME: inline this **/  
const char * flib_key::username() const 
{
	return user->name(); 
}

