/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
 *
 * net/resolver.cc
 * (c) 2005 Murat Deligonul
 */
 
#include "autoconf.h"

#include <algorithm>
#include <iterator>
#include <csignal>
#include <cstdio>
#include <cstring>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#ifdef HAVE_SYS_FILIO_H
#   include <sys/filio.h>
#endif
#ifdef HAVE_SYS_IOCTL_H
#   include <sys/ioctl.h>
#endif
#include "util/strings.h"
#include "net/resolver.h"

#include "debug.h"

namespace net {

using namespace util::strings;

/* static */ const struct resolver::resolver_thread 
	resolver::default_rt = { 0, 0 };

		
resolver::resolver(unsigned max_threads, unsigned max_lookups) 
	: 	MAX_THREADS(max_threads), 
		MAX_LOOKUPS_PER_THREAD(max_lookups),
		current_id(0),
		active_threads(0),		
		first_thread(default_rt)
{
	assert(max_threads > 0);
	assert(max_lookups > 0);
	
	/** 
	  * Set up:
	  * -- FIFO for relaying results
	  * -- mutex and condition variables 
	  */
	if (pipe(fifo) < 0) {
		// FIXME: complain
		abort();
	}
		
	pthread_mutex_init(&queue_mutex, 0);
	pthread_mutex_init(&threads_mutex, 0);
	pthread_cond_init(&queue_cv, 0);
	
	DEBUG("resolver::resolver() : [%p] READY w/ fifo: [%d:%d] max threads: %d max lpt: %d\n",
				this, fifo[0], fifo[1], max_threads, max_lookups);
}
		
/* Destructor */
resolver::~resolver()
{
	if (first_thread.self) {
		pthread_kill(first_thread.thread, SIGKILL);
	}
	pthread_mutex_destroy(&queue_mutex);
	pthread_mutex_destroy(&threads_mutex);
	pthread_cond_destroy(&queue_cv);
	
	close(fifo_reader_fd());
	close(fifo_writer_fd());
}

/* static */ resolver::request * resolver::create_request(int family,
				const char * hostname,
				unsigned short port,
				int options,
				resolver_callback * callback) 
{
	struct request * req = new request;
	req->id 	= 0;		/* no ID is assigned until put into wait queue */
	req->family 	= family;
	req->options	= options;
	req->port	= port;
	req->ai		= 0;
	req->callback	= callback;
	my_strlcpy(req->name, hostname, sizeof(req->name));
	return req;
}


/**
 * Return ID of new async request on success.
 * Otherwise return error code 
 **/
int resolver::async_lookup(resolver::request * req)
{
	req->id =  ++current_id;
	/* we reach this point with a lookup ready to go */
	pthread_mutex_lock(&queue_mutex);
	request_queue.push(req);
	unsigned int n = request_queue.size();
	pthread_mutex_unlock(&queue_mutex);
	
	/* determine if we have to dispatch a new thread */
	if (active_threads < MAX_THREADS 
		&& n > (active_threads * MAX_LOOKUPS_PER_THREAD))
	{
		resolver_thread * rt;
		if (!first_thread.self) 
			rt = &first_thread; 
		else
			rt = new resolver_thread;
		rt->self = this;
		
		pthread_attr_t attr;
		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	
		pthread_mutex_lock(&threads_mutex);	
		++active_threads;
		pthread_mutex_unlock(&threads_mutex);
		
		/* race condition ?? */
		n = pthread_create(&rt->thread, &attr, resolver_thread_fn, rt);
		pthread_attr_destroy(&attr);
	}
	
	pthread_cond_broadcast(&queue_cv);
	
	return req->id;
}


/**
 * Put a request in the cancellation queue to ignore
 * the result of request 'i' when we come across it .. 
 */
int resolver::cancel_async_lookup(int i)
{
	cancelled_requests.push_back(i);
	return cancelled_requests.size();
}

/* static */ void * resolver::resolver_thread_fn(void * param)
{
	assert(param);
	resolver_thread * rt = (resolver_thread *) param;
	resolver * self = rt->self;
	request * req = NULL;
	bool first = (rt == &self->first_thread);
	
	DEBUG("resolver::resolver_thread_fn() -- new thread\n");
	
	pthread_mutex_lock(&self->queue_mutex);
	
	while (first || self->request_queue.size() > self->MAX_LOOKUPS_PER_THREAD)
	{
		while (self->request_queue.empty())
			pthread_cond_wait(&self->queue_cv, &self->queue_mutex);
		req = self->request_queue.front();
		self->request_queue.pop();
		pthread_mutex_unlock(&self->queue_mutex);
		
		DEBUG("resolver::resolver_thread_fn() -- processing request: %d\n", req->id);
		/** perform the blocking lookup **/
		self->process_request(req);
		self->write_result(req);
		
		pthread_mutex_lock(&self->queue_mutex);
	}
	
	pthread_mutex_unlock(&self->queue_mutex);
	
	/* No more to do. If we're not the first thread, delete 
	 * our data parameter and kill thread */
	pthread_mutex_lock(&self->threads_mutex);
	--self->active_threads;
	pthread_mutex_unlock(&self->threads_mutex);
	
	DEBUG("resolver::resolver_thread_fn(): Thread exiting\n");
	if (!first)
		delete rt;
	pthread_exit(0);
	return 0; 	/* NOT REACHED */
}


int resolver::process_request(resolver::request * req)
{
	if (req->options & OPT_REVERSE_LOOKUP) {
		/* call this first to fill addrinfo field. downside is we will
		 * wait a while if the 'IP address' was invalid. */
		resolve_address(req->family, req->name, req->port, &req->ai);
		req->iresult = reverse_lookup(req->name, req->name, sizeof(req->name));
	}
	else { 
		req->iresult = resolve_address(req->family, req->name, req->port, &req->ai);
	}
	return req->iresult;		
}

/** 
 * Write the result to FIFO to be read by main thread.
 * Currently we are cheap and just write the pointer to the request
 * data structure. 
 * This operation completes the work in the resolving thread.
 *
 * FIXME: does the FIFO need to be protected with a lock? I don't think so,
 *  	  hopefully not ..
 */
int resolver::write_result(const resolver::request * req)
{
	return write(fifo_writer_fd(), &req, sizeof(resolver::request *));
}


/** Read async look-up results **/
int resolver::process_results()
{
	int processed = 0;
	int i = 0;	
	do {
		// FIXME: why is this in the loop?
		ioctl(fifo_reader_fd(), FIONREAD, &i);
		if (unsigned(i) >= sizeof(struct request *)) {
			struct request * req = read_result();
			process_result(req);
			++processed;
			/* at this point the request has resolved,
			 * the callback function notified, and any allocated
			 * resources for the request freed */
		}
		i -= sizeof(struct request *);
	} while (i > 0 && unsigned(i) >= sizeof(struct request *));
	
	return processed;
}

/** read an indiviual result from the FIFO */
resolver::request * resolver::read_result()
{
	request * req = NULL;
	read(fifo_reader_fd(), &req, sizeof(req));
	return req;
}

int resolver::process_result(request * req)
{
	assert(req);
	assert(req->id > 0);
	
	bool cancelled = false;
	
	do {
		/* First check if it's been cancelled */
		std::vector<int>::iterator i = 
			std::find(cancelled_requests.begin(),
			cancelled_requests.end(),
			req->id);
	
		if (i != cancelled_requests.end()) {
			DEBUG("resolver::process_result(): CANCELLED: %d\n", req->id);
			/* Must cancel this one */
			cancelled_requests.erase(i);
			cancelled = true;
			continue;
		}
	} while (0);	/* we must get rid of potential duplicate cancellation requests */
	
	if (!cancelled) {
		if (req->iresult == 0) {
			req->callback->async_lookup_finished(req);
		} 
		else {
			req->callback->async_lookup_failed(req);
		}
	}
	
	delete req;
	return 0;
}


/** 
 * [BLOCKING] The unified address resolver
 * handles both IPv4 and IPv6 address families 
 *
 * Returns: [same as getaddrinfo()]
 * 		0:		success
 *		<> 0:		errors
 */ 
/* static */ int resolver::resolve_address(int family,
			const char *hostname,
			unsigned short port,
			struct addrinfo ** pres)
{
	struct addrinfo hints;
	char  portbuf[6] = "";
	
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = family;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	
	sprintf(portbuf, "%u", port);
	
	return getaddrinfo(hostname, portbuf, &hints, pres);
}

/** 
 * [BLOCKING] 
 * IP->hostname, store in buffer,
 * Return 0 on success, otherwise getnameinfo() error code
 * Ugly, but it works. 
 */
/* static */ int resolver::reverse_lookup(const char * hostname, 
			char * buffer, 
			size_t len)
{
	sockaddr_in6 sin;
	int size;
	memset(&sin, 0, sizeof(sin));
	sin.sin6_family = AF_INET6;
	size =
#ifdef SIN6_LEN   /* Linux doesn't have this (?) */
	sin.sin6_len =
#endif
		sizeof(sin);

	if (inet_pton(AF_INET6, hostname, (void *) &sin.sin6_addr) < 1) {
		struct sockaddr_in * ps = (sockaddr_in *) &sin;
        	memset(&sin, 0, sizeof(sin));
        	if (inet_pton(AF_INET, hostname,
			(void *) &ps->sin_addr) < 1)
			return EAI_FAMILY;
		ps->sin_family  = AF_INET;
		size =
#ifdef SIN6_LEN     /* Linux doesn't have this (?) */
		ps->sin_len =
#endif
			sizeof(sockaddr_in);
	}
	
	return getnameinfo((const sockaddr *) &sin, size,
		buffer, len, NULL, 0, 0);		//  used to have NI_NAMEREQD ?
}


/** 
 * Tests if binding to this interface will work. Return:
 *  0  - Success
 *  1  - Failed, check errno for reason
 * -1  - name lookup failure
 *
 */
/* static */ int resolver::test_bind(const char * host, unsigned short port)
{
	struct addrinfo * ai = NULL;
	if (resolve_address(PF_UNSPEC, host, port, &ai) != 0)
		return -1;

	int s, ret = 0;
	
	/* test bind */
	s = socket(ai->ai_family, SOCK_STREAM, 0);
	if (s < 0 || bind(s, ai->ai_addr, ai->ai_addrlen) < 0) {
		ret = 1;
	}
	
	/* works */	
	close(s);
	freeaddrinfo(ai);
	return ret;
}


int resolver::lookup(const char * host, 
			char * buffer, 
			size_t len)
{
	struct addrinfo * ai = NULL;
	int i = resolve_address(PF_UNSPEC, host, 0, &ai);
	if (i != 0) {
		return i;
	}
		
	void * raw = (ai->ai_family == AF_INET6) ?
		(void *) &((sockaddr_in6 *) ai->ai_addr)->sin6_addr :
		(void *) &((sockaddr_in *) ai->ai_addr)->sin_addr;
	
	inet_ntop(ai->ai_family, raw, buffer, len);
	freeaddrinfo(ai);
	return 0;
}


/**
 * Convert a raw numeric IP address into ascii representation ..
 *  raw  -- the raw address in network order
 *  size, buff, buff_size -- ditto
 * return:
 * -1 -- failure
 *  otherwise -- success
 */
/* static */ int resolver::raw_to_ip(const struct sockaddr * raw, 
				size_t size, 
				char * buff, 
				size_t buff_size)
{
	assert(raw && buff);
	struct sockaddr_in sin;
	struct sockaddr_in6 sin6;
	
	if (size >= sizeof(sin6) ) /* IPv6 address */
	{
		memset(&sin6, 0,   sizeof sin6);
		memcpy(&sin6, raw, sizeof sin6);
		if (inet_ntop(AF_INET6, &sin6.sin6_addr, buff, buff_size)) {
			return sizeof sin6;  /* success */
		}
	}
	if (size >= sizeof(sin))  /* IPv4 address */
	{
		memset(&sin, 0,   sizeof sin);
		memcpy(&sin, raw, sizeof sin);
		if (inet_ntop(AF_INET, &sin.sin_addr, buff, buff_size)) {			
			return sizeof(sin);
		}
	}
	return -1;
}


/**  
 * Return IP address as raw data:
 * 	sockaddr_in structure required for AF_INET
 * 	sockaddr_in6 structure required for AF_INET6
 * data is in network order. 
 **/
int resolver::ip_to_raw(const char * addr, 
			struct sockaddr * buff, 
			size_t size)
{
	assert(addr && buff && size);
	struct sockaddr_in sin;
	memset(&sin, 0, sizeof(sin));
	
	if (inet_pton(AF_INET, addr, &sin.sin_addr) > 0) {
		if (size < sizeof(sin)) {
			return -1;
		}
		memcpy(buff, &sin, size);
		return sizeof(sin);
	}
	
	/* didn't work. try ipv6 */
	struct sockaddr_in6 sin6;
	memset(&sin6, 0, sizeof(sin6));
	
	if (inet_pton(AF_INET6, addr, &sin6.sin6_addr) > 0) {
		if (size < sizeof(sin6)) {
			return -1;
		}
		memcpy(buff, &sin6, sizeof(sin6));
		return sizeof(sin6);
	}
	return -1;
}

/* static */ int resolver::is_ip_address(const char * addr)
{
	char buff[sizeof(struct sockaddr_in6)];
	return (ip_to_raw(addr, (sockaddr *) buff, sizeof(buff)) > 0);
}


/* static */ char * resolver::result_to_string(const resolver::request * req, 
						char * buf, 
						size_t  len)
{
	if (req->options & OPT_REVERSE_LOOKUP) {
		my_strlcpy(buf, req->name, len);
	}
	else {
		raw_to_ip(req->ai->ai_addr, req->ai->ai_addrlen, buf, len);
	}
	return buf;
}				

} /* namespace net */
